/*
 * JSON Stream Parser
 * Copyright 2010,2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of the JSON Stream Parser.
 *
 * JSON Stream Parser is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JSON Stream Parser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JSON Stream Parser.  If not, see <http://www.gnu.org/licenses/>.
 */

package l.joe.json;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.ParseException;
import java.util.Locale;
import java.util.Stack;

/**
 * @author Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 *
 */
public class JSONParser {

    private static enum State { VALUE, OBJECT, ARRAY };
    private static final DecimalFormat JSON_DECIMAL = new DecimalFormat("0.#E0",
            DecimalFormatSymbols.getInstance(Locale.ENGLISH));

    private JSONHandler handler;

    private StringBuffer sb = new StringBuffer();
    private int cached=-1;
    private Stack<State> stack = new Stack<State>();

    private JSONParser(JSONHandler handler)
    {
        this.handler = handler;
    }

    public static void parse(InputStream in, JSONHandler handler) throws JSONParserException
    {
        Reader isr = null;
        try
        {
            isr = new InputStreamReader(in, "UTF-8");
            new JSONParser(handler).parse(isr);
            isr.close();
        } 
        catch (IOException e)
        {
            throw new JSONParserException(e);
        }
        finally 
        {
            if (isr!=null) try { isr.close(); } catch (IOException e) {};
        }
    }

    private void parse(Reader isr) throws IOException, JSONParserException
    {
        int c;
        State state = State.VALUE;
        while ((c = readNextNonWhitespace(isr)) >= 0)
        {
            switch (state) {
                case OBJECT:
                    switch (c) {
                        case '}' : handler.endObject(); state=popPeekStack(); break; 
                        case '\"': 
                            handler.objectValueName(readString(isr));
                            if (readNextNonWhitespace(isr)!= ':') throw new JSONParserException("Missing ':' after object property name!");
                            state = State.VALUE;  
                            break;
                        case ',' : //TODO also allows , before first value
                            //nothing to do
                            break;
                        default : new JSONParserException("Expecting start of String or comma, found '"+(char) c+"'!");
                    } break;
                case ARRAY:
                    switch (c) {
                        case ']' : handler.endArray(); state=popPeekStack(); break;
                        case ',' : state=State.VALUE; break;
                    } break;
                case VALUE:
                    switch (c) {
                        case '{' : handler.startObject(); state=stack.push(State.OBJECT); break; 
                        case '[' : handler.startArray(); stack.push(State.ARRAY); break; 
                        case ']' : handler.endArray(); state=popPeekStack(); break;
                        case '\"': handler.stringValue(readString(isr)); state=peekStack(); break; 
                        case 't': readSequence(isr, 'r','u','e'); handler.booleanValue(true); state=peekStack(); break; 
                        case 'f': readSequence(isr, 'a','l','s','e'); handler.booleanValue(false); state=peekStack(); break; 
                        case 'n': readSequence(isr, 'u', 'l', 'l'); handler.nullValue(); state=peekStack(); break; 
                        //number
                        default:
                            handler.numberValue(readNumber(c,isr)); state=peekStack(); break;
                    } break;
                default:
                    throw new JSONParserException("undefined parser state");
            }
        }
    }

    private State peekStack()
    {
        //avoid empty stack exception
        if (stack.isEmpty()) return null;
        return stack.peek();
    }

    private State popPeekStack()
    {
        stack.pop();
        return peekStack();
    }

    private int readNextNonWhitespace(Reader isr) throws IOException
    {
        int c=-1;

        if (cached!=-1) {
            c = cached; cached=-1; return c;
        }

        while ((c=isr.read())>=0 && Character.isWhitespace(c));

        return c;
    }

    private String readString(Reader isr) throws IOException, JSONParserException
    {
        if (sb.length()!=0) sb.delete(0, sb.length());

        int c;
        while ((c=isr.read())!='\"' && c>=0)
        switch (c) {
            case '\\': 
                c=isr.read();
                switch (c) {
                    case 'b': sb.append('\b'); break;
                    case 'f': sb.append('\f'); break;
                    case 'n': sb.append('\n'); break;
                    case 'r': sb.append('\r'); break;
                    case 't': sb.append('\t'); break;
                    case '/': sb.append('/');  break;
                    case '\"': sb.append('\"'); break;
                    case '\\': sb.append('\\'); break;
                    case 'u' :
                        c=0;
                        for (int i=0; i<4; i++)
                            c= (c << 4) + hexValue(isr.read());
                        sb.append((char) c);
                        break;
                    default: throw new JSONParserException("Wrong escape sequence in String: '\\"+(char) c +"'");
                } break;
            default: sb.append((char) c);
        }

        return sb.toString();
    }

    private int hexValue(int i) throws JSONParserException
    {
        if (i==-1) throw new JSONParserException("Unexpected end of Stream!");

        if ('0'<=i && i <='9') return i - '0';
        else if ('a'<=i && i <='f') return i - 'a' + 10;
        else if ('A'<=i && i <='F') return i - 'A' + 10;

        throw new JSONParserException("Invalid hex value in unicode escape '"+(char) i+"'");
    }

    private void readSequence(Reader isr, char... sequence) throws IOException, JSONParserException
    {
        int c;
        for (int i=0; i < sequence.length; i++)
            if ((c=isr.read())!=sequence[i]) throw new JSONParserException("Invalid value, expected one of {true, false, null}");
    }

    private Number readNumber(int c, Reader isr) throws IOException, JSONParserException
    {
        sb.delete(0, sb.length());
        
        boolean stop = false;
        do switch (c) {
                case '1': case '2': case '3': case '4': case '5':
                case '6': case '7': case '8': case '9': case '0':
                case 'E': case '.': case '-': case '+':
                    sb.append((char)c);
                    break;
                case 'e':
                    sb.append('E');
                    break;
                // object, array ended, or next value
                case '}':
                case ']':
                case ',':
                    cached=c;
                //whitespace or something invalid
                default:
                    if (sb.length()==0) throw new JSONParserException("Invalid number");
                    stop=true;
                    break;
            } while(!stop && (c=isr.read())>=0);

        Number number = null;
        try {
            number = JSON_DECIMAL.parse(sb.toString());
        } catch (ParseException e) {
            throw new JSONParserException("Invalid number: " + sb.toString(),e);
        }

        return number;
    }

}
