/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.core.config;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.UIManager;

import org.kuehntopf.streamtastic.utils.Utilities;

/**
 * Configuration.
 * This class reads and writes properties
 * to a special configuration file. It is used
 * for storing all user preferences
 * @author akuehntopf
 */
public class Config
{
    public static enum Key {
        // Streamripper
        STREAMRIPPER_CMD, STREAMRIPPER_OPTS, STREAMRIPPER_RELAY_PORT_RANGE, DESTINATION_DIR,
        // Shoutcast
        SHOUTCAST_MAX_LIST_LENGTH,
        // Icecast
        ICECAST_GENRE_BLACKLIST,
        // external Player
        PLAYER_CMD, PLAYER_OPTS,
        // GUI settings
        WINDOW_STATE, DIR_STATION_DOUBLE_CLICK_ACTION, FAVS_STATION_DOUBLE_CLICK_ACTION,
        PREF_SHOW_ADVANCED_TAB, MINIMIZE_TO_TRAY, LOOK_AND_FEEL;

        private static Map<Key, ParameterInfo> parameterDetails = new HashMap<Key, ParameterInfo>();
        static {
            initParameterInfos(parameterDetails);
        }

        public ParameterInfo getParameterInfo() {
            return parameterDetails.get(this);
        }
    }

    private static final Map<String, String> customLAFNameToClassName = new TreeMap<String, String>();
    static 
    {
        customLAFNameToClassName.put("SUBSTANCE", "org.pushingpixels.substance.api.skin.SubstanceNebulaLookAndFeel");
        customLAFNameToClassName.put("OS", UIManager.getSystemLookAndFeelClassName());
        customLAFNameToClassName.put("JAVA", UIManager.getCrossPlatformLookAndFeelClassName());
    }
    
    private static final String DEFAULT_CONFIG = "/org/kuehntopf/streamtastic/resources/config/default.properties";
    private static  Config        _instance = null;
    private         Properties    _props, defaultConfig;
    private         String        _fileName = Utilities.getStorageDirectory() + 
            File.separator + "config.properties";
    private PropertyChangeSupport pcs = new PropertyChangeSupport(this);
    private Set<Key> validated = new TreeSet<Key>();
    
    /**
     * Returns the Singleton instance
     * of this class
     */
    public static Config 
    getInstance()
    {
        if (_instance == null)
            _instance = new Config();
        return _instance;
    }
    
    /**
     * Constructs a new ConfigManager object.
     * This is private for Singleton reasons
     */
    public 
    Config()
    {
        defaultConfig = new Properties();
        try
        {
            InputStream defaultsIn = Config.class.getResourceAsStream(DEFAULT_CONFIG);
            defaultConfig.load(defaultsIn);
            defaultsIn.close();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        };
        _props = new Properties(defaultConfig);
        
        File propFile = new File(_fileName);
        
        // Create Configuration file if it
        // doesn't exist
        if (!propFile.exists())
        {
            try
            {
                propFile.createNewFile();
            } catch (IOException ex)
            {
                ex.printStackTrace();
            }
        }
        
        // Finally try loading properties
        loadProperties();
    }
    
    public void 
    addPropertyChangeListener(PropertyChangeListener listener)
    {
        pcs.addPropertyChangeListener(listener);
    }

    public void 
    addPropertyChangeListener(Key property, PropertyChangeListener listener)
    {
        pcs.addPropertyChangeListener(property.toString(), listener);
    }

    public void 
    removePropertyChangeListener(PropertyChangeListener listener)
    {
        pcs.removePropertyChangeListener(listener);
    }

    public void 
    removePropertyChangeListener(Key property, PropertyChangeListener listener)
    {
        pcs.addPropertyChangeListener(property.toString(), listener);
    }

    /**
     * Sets a new Property value
     * @param key property key to set
     * @param value the value to set
     */
    public void 
    setProperty(Key key, String value)
    throws IllegalArgumentException
    {
        if (!key.getParameterInfo().isValid(value))
            throw new IllegalArgumentException("Value not valid!");

        String oldValue = getProperty(key);
        _props.remove(key.toString());
        if (!_props.getProperty(key.toString()).equals(value))
            _props.setProperty(key.toString(), value);
        validated.add(key);
        saveProperties();
        pcs.firePropertyChange(key.toString(), oldValue, value);
    }
    
    /**
     * Gets a Property value
     * @param key key to look for
     * @return the property value
     */
    public String 
    getProperty(Key key)
    {
        String value = _props.getProperty(key.toString());
        if (!validated.contains(key))
        {
            if (key.getParameterInfo().isValid(value))
                validated.add(key);
            else
            {
                _props.remove(key.toString());
                value = _props.getProperty(key.toString());
                // check whether default value is valid
                if (key.getParameterInfo().isValid(value))
                    validated.add(key);
            }
        }
        return value != null && value.trim().length() == 0 ? null : value;
    }  
           
    public boolean 
    isValueValidForKey(Key key) 
    {
        getProperty(key);
        return validated.contains(key);
    }

    public boolean
    isDefaultValue(Key target)
    {
        return hasDefaultValue(target)
                && _props.getProperty(target.toString()) == defaultConfig.getProperty(target
                        .toString());
    }

    public void
    setToDefault(Key target)
    {
        if (!hasDefaultValue(target)) return;

        String oldValue = getProperty(target);
        _props.remove(target.toString());
        validated.remove(target);
        saveProperties();
        pcs.firePropertyChange(target.toString(), oldValue, getProperty(target));
    }

    public boolean 
    hasDefaultValue(Key target)
    {
        return defaultConfig.getProperty(target.toString()).length() != 0;
    }

    public String
    getDefaultValue(Key target)
    {
        String defaultValue = defaultConfig.getProperty(target.toString());
        return defaultValue.length() != 0 ? defaultValue : null;
    }

    /**
     * Saves the properties
     * to disk.
     */
    public void 
    saveProperties()
    {
        try
        {
            File propsFile = new File(_fileName);
            _props.store(new FileOutputStream(propsFile), "Streamtastic");
        } catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    
    /**
     * Loads the Properties from
     * disk.
     */
    public void 
    loadProperties()
    {
        try
        {
            _props.load(new FileInputStream(new File(_fileName)));
        } catch (Exception ex)
        {            
            ex.printStackTrace();
        }
    }  
    
    public int[] 
    getStreamRipperRelayPortRange() 
    {
        String portrange = Config.getInstance().getProperty(Key.STREAMRIPPER_RELAY_PORT_RANGE);
        int[] ports = null;
        if (portrange != null) 
        {
            String[] portStrings = portrange.split("-");
            if (portStrings.length==2) 
            {
                ports= new int[2];
                for (int i=0; i < ports.length; i++) try
                {
                    ports[i] = Integer.valueOf(portStrings[i].trim());
                }
                catch (NumberFormatException nfe) 
                {
                    System.err.println("STREAMRIPPER_RELAY_PORT_RANGE is set to an invalid value!");
                    ports=null;
                }
            }
        }
        
        return ports;
    }

    public String
    getLookAndFeelClassName()
    {
        return getLookAndFeelClassName(getProperty(Key.LOOK_AND_FEEL));
    }

    public String 
    getLookAndFeelClassName(String lafValue)
    {
        return customLAFNameToClassName.containsKey(lafValue)? customLAFNameToClassName.get(lafValue) : lafValue;
    }

    public WindowState
    getWindowState()
    {
        try
        {
            return new WindowState(getProperty(Key.WINDOW_STATE));
        }
        catch (IllegalArgumentException e)
        {
            return null;
        }
    }

    private static void initParameterInfos(Map infos)
    {
        BufferedReader in = new BufferedReader(new InputStreamReader(Config.class
                .getResourceAsStream(DEFAULT_CONFIG)));
        Pattern parameter = Pattern.compile("\\s*([^#][^=]*)\\s*=.*");
        Pattern parameterConfig = Pattern
                .compile("\\s*#\\s*PARAMETER-CONFIG\\s*"
                        + "\\[\\s*([^\\[;\\s]*?)\\s*(?:\\[\\s*(.*?)\\s*\\])?\\s*(?:;\\s*([\\w.]*)\\s*)?\\].*");
        try
        {
            String currentLine = null;
            Key currentKey = null;
            while ((currentLine = in.readLine()) != null)
            {
                Matcher matcher = parameter.matcher(currentLine);
                if (matcher.matches())
                {
                    currentKey = Key.valueOf(matcher.group(1));
                    continue;
                }

                matcher = parameterConfig.matcher(currentLine);
                if (matcher.matches())
                {
                    if (currentKey == null)
                    {
                        System.err.println("No Parameter affected by: \"" + currentLine + "\"!");
                        continue;
                    }
                    infos.put(currentKey, new ParameterInfoImpl(currentKey, ParameterInfo.Type
                            .valueOf(matcher.group(1)), matcher.group(2), matcher.group(3)));
                    currentKey = null;
                }
            }
            in.close();
        }
        catch (IOException e)
        {
            System.err.println("Error intializing ParameterInfos!");
        }
    }
}

class ParameterInfoImpl implements ParameterInfo {
    private Type type;
    private Config.Key key;
    private TreeSet<String> values = new TreeSet<String>();
    private Validator validator;

    protected ParameterInfoImpl(Config.Key key, ParameterInfo.Type type, String values,
            String validatorClassName)
    {
        this.key = key;
        this.type = type;
        if (validatorClassName != null)
            try
            {
                validator = (Validator) Class.forName(validatorClassName).newInstance();
            }
            catch (Exception e)
            {
                System.err.println("Unable to set validator for key " + key + ", Class \""
                        + validatorClassName + "\" could not be instantiated! ("
                        + e.getClass().getName() + ")");
            }
        if (values == null) 
            return;
        this.values.addAll(Arrays.asList(values.split("\\s*,\\s*")));
    }

    public String[] getFixedValues()
    {
        return values.toArray(new String[values.size()]);
    }

    public Type getParameterType()
    {
        return type;
    }

    public boolean isValid(Object value)
    {
        return (value != null && values.contains(value))
                || (validator == null ? 
                        (type==Type.CUSTOM || type==Type.FIXED_VALUES_AND_CUSTOM)
                        : validator.isValid(value));
    }
}