/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.core.icecastfeed;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;
import java.util.Vector;

import javax.swing.ImageIcon;
import javax.xml.parsers.SAXParserFactory;

import org.kuehntopf.streamtastic.core.DirectoryFeed;
import org.kuehntopf.streamtastic.core.DirectoryFeedException;
import org.kuehntopf.streamtastic.core.Station;
import org.kuehntopf.streamtastic.utils.Utilities;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * DirectoryFeed implementation for icecast directory on http://dir.xiph.org
 * 
 * @author mikeL@launchpad.net
 */
public class IcecastFeed extends DirectoryFeed {

    private final static String DIR_URL = "http://dir.xiph.org/yp.xml";
    private final static String GENRE_ALL = Utilities.getString("ICE_GENRE_ALL");
    private final static String GENRE_NO_VALUE = Utilities.getString("ICE_GENRE_OTHER");
    private static final int MAX_GENRE_SIZE_MINIMUM=10;
    private final static ImageIcon[] ICONS = new ImageIcon[] {
            Utilities.loadIcon("small/icecast.png"), Utilities.loadIcon("large/icecast.png") };

    private Map<String, List<Station>> genreToStation = new TreeMap<String, List<Station>>();
    
    // register Feed implementation
    static 
    {
        DirectoryFeed.registerDirectoryFeed(DIR_URL, IcecastFeed.class);
    }

    private synchronized void parseStations() throws DirectoryFeedException
    {
        if (!genreToStation.isEmpty())
            return;

        final Map<String, List<Station>> genreToStationMap = new TreeMap<String, List<Station>>();

        try
        {
            SAXParserFactory parserFactory = SAXParserFactory.newInstance();
            //ensure qName!=null
            parserFactory.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
            
            parserFactory.newSAXParser().parse(getInputStreamForHttpUrl(new URL(DIR_URL)),
                    new DefaultHandler() {
                
                private final static String STATION_NODE = "entry";
                
                private StringBuffer currentCData = new StringBuffer();
                private Properties stationProperties = null;
                
                @Override
                public void characters(char[] ch, int start, int length) throws SAXException
                {
                    currentCData.append(ch, start, length);
                }

                @Override
                public void endElement(String uri, String localName, String qName)
                        throws SAXException
                {
                    if (STATION_NODE.equals(qName))
                    {
                        IcecastStation station = new IcecastStation(stationProperties);
                        addStation(genreToStationMap, GENRE_ALL, station);
                        if (station.getGenres().length == 0)
                            addStation(genreToStationMap, GENRE_NO_VALUE, station);
                        else
                            for (String genre : station.getGenres())
                                addStation(genreToStationMap, genre, station);
                        
                        stationProperties = null;
                        return;
                    }
                    
                    if (stationProperties != null)
                        stationProperties.setProperty(qName, currentCData.toString());
                }

                @Override
                public void startElement(String uri, String localName, String qName,
                        Attributes attributes) throws SAXException
                {
                    if (STATION_NODE.equals(qName))
                        stationProperties = new Properties();

                    if (currentCData.length() != 0)
                        currentCData.delete(0, currentCData.length());
                }
            });

        } catch (Exception e)
        {
            throw new DirectoryFeedException(e);
        }

        compressGenres(genreToStationMap);

        genreToStation.clear();
        genreToStation.putAll(genreToStationMap);
    }

    @Override
    protected InputStream getInputStreamForHttpUrl(URL url) throws IOException
    {
        // TODO output debug messages to optional log file
        boolean debug = false;

        FileOutputStream fo = null;
        InputStream in = null;
        try
        {
            long startTime = System.currentTimeMillis();

            in = new UTF8DeOverEncodingFilterStream(super.getInputStreamForHttpUrl(url));

            if (!debug) return in;

            // in debug mode save xml data to file and parse that file afterwards
            // so we are able to distinguish why xml parsing failed
            String filename = Utilities.getHome() + File.separator + ".streamtastic"
                    + File.separator + "debug-icecast.xml";
            fo = new FileOutputStream(filename);
            byte[] buffer = new byte[1024];
            int length;
            while ((length = in.read(buffer))!=-1) 
                fo.write(buffer, 0, length);
            fo.close();
            in.close();

            System.out.println("\n>>> Time: " + ((System.currentTimeMillis()-startTime)/1000f)+"s");
            return new BufferedInputStream(new FileInputStream(filename));
        }
        finally 
        {
            if (fo!=null) fo.close();
            if (in!=null && debug) in.close(); 
        }
    }

    private void addStation(Map<String, List<Station>> genreToStationMap, String genre,
            Station station)
    {
        List<Station> stationList = genreToStationMap.get(genre);
        if (stationList == null)
            stationList = new Vector<Station>(5);

        stationList.add(station);
        genreToStationMap.put(genre, stationList);
    }
    
    private void compressGenres(Map<String, List<Station>> genreToStationMap) 
    {
        // TODO output debug messages to optional log file
        boolean debugMessages = false;
        
        // cleanup sparsely populated genres
        // put stations which become homeless into GENRE_NO_VALUE
        if (debugMessages)
            System.out.println("ICECAST-GENRE-REDUCTION");
        
        Map<Integer, List<String>> sizeToGenreList = new HashMap<Integer, List<String>>();
        for (String key : genreToStationMap.keySet())
        {
            if (key == GENRE_ALL || key == GENRE_NO_VALUE)
                continue;
            int size = genreToStationMap.get(key).size();
            if (size >= MAX_GENRE_SIZE_MINIMUM)
                continue;
            List<String> genres = sizeToGenreList.get(size) == null ? new ArrayList<String>()
                    : sizeToGenreList.get(size);
            genres.add(key);
            sizeToGenreList.put(size, genres);
        }

        int genresCountBefore = genreToStationMap.keySet().size();
        int genresCountAfter = genresCountBefore;
        int homelessStationsCount = genreToStationMap.get(GENRE_NO_VALUE) == null ? 0
                : genreToStationMap.get(GENRE_NO_VALUE).size();
        int genreReduction = genresCountBefore - genresCountAfter;
        int previousGenreReduction, previousHomlessStationsCount;

        for (int i = 1; i < MAX_GENRE_SIZE_MINIMUM; i++)
        {
            if (sizeToGenreList.get(i)==null) continue;

            for (String key : sizeToGenreList.get(i))
            {
                List<Station> stations = genreToStationMap.get(key);
                genreToStationMap.remove(key);
                for (Station station : stations)
                    if (((IcecastStation) station).removeGenre(key) == 0)
                        addStation(genreToStationMap, GENRE_NO_VALUE, station);
            }

            previousGenreReduction = genreReduction;
            genresCountAfter = genreToStationMap.keySet().size();
            previousHomlessStationsCount = homelessStationsCount;
            homelessStationsCount = genreToStationMap.get(GENRE_NO_VALUE) == null ? 0
                    : genreToStationMap.get(GENRE_NO_VALUE).size();
            genreReduction = genresCountBefore - genresCountAfter;

            if (debugMessages)
                System.out.print("  min. " + (i + 1) + " entries: " + genresCountBefore + "=>"
                    + genresCountAfter + " thus " + genreReduction + "(+"
                    + (genreReduction - previousGenreReduction) + ") removed, resulting in "
                    + homelessStationsCount + "(+"
                    + (homelessStationsCount - previousHomlessStationsCount)
                    + ") homeless stations (" + (genreReduction - previousGenreReduction) + ":"
                    + (homelessStationsCount - previousHomlessStationsCount) + "="
                    + (genreReduction - previousGenreReduction)
                    / (float) (homelessStationsCount - previousHomlessStationsCount) + ") ");

            if ((genreReduction - previousGenreReduction)
                    - (homelessStationsCount - previousHomlessStationsCount) < 0)
            {
                if (debugMessages)
                    System.out.println("=> ABORT\n");
                break;
            }
            if (debugMessages)
                System.out.println("=> CONTINUE");
        }
    }

    @Override
    public ImageIcon[] getImageIcons()
    {
        return ICONS;
    }

    @Override
    public String getName()
    {
        return Utilities.getString("SIDEBARMENU_ICECAST");
    }

    public List<String> getGenres() throws DirectoryFeedException
    {
        parseStations();
        return new ArrayList<String>(genreToStation.keySet());
    }

    /**
     * Returns a list of stations for the given search term
     * 
     * @param search
     *            term to list
     * @return a list of stations
     */

    public List<Station> getSearchResult(String searchTerm, Integer bitrate)
            throws DirectoryFeedException
    {
        parseStations();
        int br = bitrate == null ? -1 : bitrate;

        return filter(genreToStation.get(GENRE_ALL), br, searchTerm);
    }

    public List<Station> getStations(String genre, Integer bitrate) throws DirectoryFeedException
    {
        parseStations();
        int br = bitrate == null ? -1 : bitrate;

        return filter(genreToStation.get(genre), br, null);
    }

    @Override
    public void flush()
    {
        synchronized (genreToStation)
        {
            genreToStation.clear();
        }
    }

    private List<Station> filter(List<Station> list, int br, String query)
    {
        if (list == null) // prevent NullpointerException
            return new Vector<Station>(1);
        if (br == -1 && query == null)
            return list;
        String st = query == null ? null : query.toLowerCase();

        List<Station> filteredList = new Vector<Station>(list.size() / 2);
        for (Station station : list)
        {
            int stabr = station.getBitrate() == null ? -1 : station.getBitrate();
            if ((br == -1 || stabr == br) && // Bitrate filtern
                    (st == null || stationIsMatch(station, st)))
                filteredList.add(station);
        }

        return filteredList;
    }

    private boolean stationIsMatch(Station station, String query)
    {
        for (String term: query.split("\\s"))
            if (term.length() > 0 && !station.getName().toLowerCase().contains(term)
                                  && !station.getCurrentTrack().toLowerCase().contains(term)
                                  && !station.getGenre().toLowerCase().contains(term))
                return false;

        return true;
    }

    @Override
    public Station.Property[] getSupportedStationProperties()
    {
        return new Station.Property[] { Station.Property.NAME, Station.Property.CURRENT_TRACK, Station.Property.BITRATE, 
                                        Station.Property.MIMETYPE, Station.Property.GENRE,  Station.Property.URL };
    }

}

class UTF8DeOverEncodingFilterStream extends FilterInputStream {

    // buffer great enough for 5 times applied utf8 encoding
    private int[] buffer = new int[(2 << 5) + 1];
    private int cached = -1;
    private int currentlength = 0, writeout = 0;

    private static final int[] VALID = new int[] { 0x82, 0x83, 0xc2, 0xc3 };
    private static final int[] C2_INSTEAD_OF_C3 = new int[] { 0xb4 };

    public UTF8DeOverEncodingFilterStream(InputStream in)
    {
        super(in);
    }

    @Override
    public int read() throws IOException
    {
        // This algorithm fixes utf8 encodings of charactes with utf8 encoding
        // in range of [c3 80 - c3 bf] created by applying utf8 too many times.
        // recovers f.i. german umlauts
        // NOTE:however backward utf8 unencoding does not work proberly
        // because often control chars like [c2 82,c2 83] are missing
        // so sometimes the char should have started with c2 instead of c3

        if (writeout < currentlength)
            return buffer[writeout++];

        int c = (cached != -1) ? cached : in.read();
        cached = -1;
        if (c != 0xc3)
            return c;

        currentlength = writeout = 0;
        while (Arrays.binarySearch(VALID, c) >= 0)
        {
            buffer[currentlength++] = c;

            // buffer is full stop and write out buffer
            if (currentlength == buffer.length)
                return buffer[writeout++];

            c = in.read();
        }

        // set the second byte
        if (0x80 <= c && c <= 0xbf)
        {
            buffer[1] = c;
            currentlength = 2;
        }
        // 0x82 or 0x83 if c does not fit
        else if (0x80 <= buffer[currentlength - 1] && buffer[currentlength - 1] <= 0xbf)
        {
            buffer[1] = buffer[currentlength - 1];
            currentlength = 2;
            cached = c;
        }

        // something other... cache c and write out the buffer
        else 
        {
            cached = c;
            return buffer[writeout++];
        }

        //decide whether to output c2 or c3 as first byte.
        if (Arrays.binarySearch(C2_INSTEAD_OF_C3, buffer[1])>=0)
            buffer[0]=0xc2;

        return buffer[writeout++];
    }

    @Override
    public int read(byte[] b, int off, int len) throws IOException
    {
        for (int i = off; i < Math.min(b.length, off + len); i++)
        {
            int c = read();
            if (c == -1)
                return (i == off) ? -1 : i - off;
            b[i] = (byte) c;
        }
        return Math.min(len, b.length - off);
    }
}
