/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.core.shoutcastfeed;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.List;
import java.util.TreeSet;

import javax.swing.ImageIcon;

import l.joe.json.JSONHandler;
import l.joe.json.JSONParser;

import org.kuehntopf.streamtastic.core.DirectoryFeed;
import org.kuehntopf.streamtastic.core.DirectoryFeedException;
import org.kuehntopf.streamtastic.core.Station;
import org.kuehntopf.streamtastic.core.config.Config;
import org.kuehntopf.streamtastic.utils.Utilities;

/**
 * Shoutcast Yellow pages class
 * This downloads genres from the shoutcast web service
 * and can also get a list of stations for a specific genre.
 * @author akuehntopf
 */
public class ShoutcastFeed extends DirectoryFeed
{
    private static final ImageIcon[] ICONS = new ImageIcon[] {
            Utilities.loadIcon("small/shoutcast.png"), Utilities.loadIcon("large/shoutcast.png") };
    private static final String B_URL = "http://winamp.shoutcast.com/embed_module";
    private static final String S_URL = B_URL + "?reqType=GENRE_SEARCH&s=";
    private static final String F_URL = B_URL + "?reqType=KEYWORD_SEARCH&s="; 
    
    private ArrayList<String> genres;
    private ArrayList<Station> stationsCache = new ArrayList<Station>();
    private String cachedURL = null;

    // register Feed implementation
    static 
    {
        DirectoryFeed.registerDirectoryFeed(B_URL, ShoutcastFeed.class);
    }

    @Override
    public ImageIcon[]
    getImageIcons()
    {
        return ICONS;
    }

    @Override
    public String
    getName()
    {
        return Utilities.getString("SIDEBARMENU_SHOUTCAST");
    }

    /**
     * Returns a list of 
     * all registered genres from the Shoutcast
     * Yellowpages
     * @return a list of genres
     */
    public List<String> 
    getGenres() 
    throws DirectoryFeedException
    {
        if (genres!=null) return genres;

        try
        {
            BufferedReader reader = new BufferedReader(new InputStreamReader(getClass()
                    .getResourceAsStream("/org/kuehntopf/streamtastic/resources/shoutcast/genres.txt")));

            TreeSet<String> set = new TreeSet<String>();
            String line;
            while ((line = reader.readLine())!=null)
                if ((line=line.trim()).length()!=0) set.add(line);
            reader.close();

            genres = new ArrayList<String>(set);
        } catch (Exception e)
        {
           throw new DirectoryFeedException(e);
        }        
        return genres;
    }

    /**
    * Returns a list of stations for
    * the given genre
    * @param genre Genre to list
    * @return a list of stations
    */
    
    public List<Station>
    getStations(String genre,Integer bitrate)
    throws DirectoryFeedException
    {
        return parseStations(S_URL, genre, bitrate == null ? null : bitrate.toString());
    }
    
     /**
     * Returns a list of stations for
     * the given search term
     * @param search term to list
     * @return a list of stations
     */
    
    public List<Station>
    getSearchResult(String searchTerm, Integer bitrate)
    throws DirectoryFeedException
    {
        return parseStations(F_URL, searchTerm, bitrate == null ? null : bitrate.toString());
    }
    
    /**
     * Returns a list of stations for
     * the given genre or search term
     * @param genre Genre to list
     * @return a list of stations
     */
    
    private List<Station>
    parseStations(String url, String querry, String bitrate) 
    throws DirectoryFeedException
    {
        String options="";
        final int intBitrate = bitrate!=null ? Integer.parseInt(bitrate) : -1;
        
        // Get list length from Config
        String maxListlength = Config.getInstance().getProperty(Config.Key.SHOUTCAST_MAX_LIST_LENGTH);
        
        // if maxListlength is set to "UNLIMITED" the whole list will be displayed
        // TODO UNLIMITED no longer possible
        if (!maxListlength.equals("UNLIMITED"))
            options +="&cnt="+maxListlength;
        else options +="&cnt="+50000;
        
        final ArrayList<Station> list = bitrate!=null ? new ArrayList<Station>() : stationsCache;

        try
        {
            String newURL = url + URLEncoder.encode(querry, "UTF-8") + options;

            if (newURL.equals(cachedURL))
            {
                if (bitrate!=null)
                {
                    for (Station station : stationsCache)
                        if (station.getBitrate()==intBitrate)
                            list.add(station);
                }

                return list;
            }

            stationsCache.clear();
            cachedURL = newURL;

            JSONParser.parse(getInputStreamForHttpUrl(new URL(newURL)), new JSONHandler() {
            
                private final static String STATIONS_ARRAY_NAME = "stations";
                private final static String STATION_ATTRIBUTE_NAME = "name";
                private final static String STATION_ATTRIBUTE_MIMETYPE = "mt";
                private final static String STATION_ATTRIBUTE_ID = "id";
                private final static String STATION_ATTRIBUTE_BITRATE = "br";
                private final static String STATION_ATTRIBUTE_GENRE = "genre";
                private final static String STATION_ATTRIBUTE_LISTENERS = "lc";
                private final static String STATION_ATTRIBUTE_TRACK = "ct";

                ShoutcastStation sta;
                boolean inStationsArray=false;
                String valueName=null;

                public void stringValue(String value)
                {
                    if (!inStationsArray) return;

                    if (STATION_ATTRIBUTE_NAME.equals(valueName)) sta.setName(value);
                    if (STATION_ATTRIBUTE_MIMETYPE.equals(valueName)) sta.setMimetype(value);
                    if (STATION_ATTRIBUTE_ID.equals(valueName)) sta.setId(Integer.parseInt(value));
                    if (STATION_ATTRIBUTE_GENRE.equals(valueName)) sta.setGenre(value);
                    if (STATION_ATTRIBUTE_TRACK.equals(valueName)) sta.setCurrentTrack(value);
                }
            
                public void startObject()
                {
                    if (!inStationsArray) return;
                    sta = new ShoutcastStation();
                }
            
                public void startArray()
                {
                    inStationsArray = STATIONS_ARRAY_NAME.equals(valueName);
                }
            
                public void objectValueName(String valueName)
                {
                    this.valueName = valueName;
                }
            
                public void numberValue(Number number)
                {
                    if (!inStationsArray) return;

                    if (STATION_ATTRIBUTE_BITRATE.equals(valueName)) sta.setBitrate(number.intValue());
                    if (STATION_ATTRIBUTE_LISTENERS.equals(valueName)) sta.setListeners(number.intValue());
                }
            
                public void endObject()
                {
                    if (!inStationsArray) return;
                    stationsCache.add(sta);
                    if (intBitrate!=-1 && intBitrate==sta.getBitrate()) list.add(sta);
                }
            
                public void endArray()
                {
                    inStationsArray=false;
                }
            
                public void nullValue() {}
                public void booleanValue(boolean value) {}
            });

        } catch (Exception e)
        {
            flush();
            throw new DirectoryFeedException(e);
        }        
        return list;
    }

    @Override
    public Station.Property[] getSupportedStationProperties()
    {
        return new Station.Property[] { Station.Property.NAME, Station.Property.CURRENT_TRACK, 
                Station.Property.LISTENERS, Station.Property.BITRATE, Station.Property.MIMETYPE, 
                Station.Property.GENRE, Station.Property.URL };
    }

    @Override
    public void flush()
    {
        cachedURL = null;
        stationsCache.clear();
    }

}