/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.core.streamripper;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;
import java.util.Observable;

import org.kuehntopf.streamtastic.core.Station;
import org.kuehntopf.streamtastic.core.config.Config;

/**
 * StreamRipper wrapper
 * @author akuehntopf
 */
public class StreamRipper extends Observable implements Runnable
{
    // Station
    private Station _station;
    // Name
    private String  _name;
    // Current Track
    private String  _track;
    // Amount transferred
    private String  _transferred;
    // Last Error
    private String  _error;
    private Integer _relayPort;
        
    private Process _proc;   
    private Thread  _myThread;
    private StreamRipperStatus _status;
    
    private static final Map<String, String> mimeTypeToExtension = new HashMap<String, String>();
    static {
        mimeTypeToExtension.put("application/ogg", ".ogg");
        mimeTypeToExtension.put("audio/aac", ".aac");
        mimeTypeToExtension.put("audio/aacp", ".aac");
        mimeTypeToExtension.put("audio/mpeg", ".mp3");
        mimeTypeToExtension.put("MP3", ".mp3");
        mimeTypeToExtension.put("video/nsv", ".nsv");
        
        // don't know what data stands for
        mimeTypeToExtension.put("data", "");
    }
    
    /**
     * Constructor
     * @param url URL to rip
     * @param targetDir Target directory to rip to
     */
    public 
    StreamRipper(Station s, Integer relayPort)
    {
        _station = s;     
        _name = s.getName();
        _relayPort = relayPort;
    }
    
    /**
     * Starts the ripping process
     * This starts a new Thread and thus
     * doesn't block.
     */
    public void 
    start()
    {
        _myThread = new Thread(this);
        _myThread.start();
    }
       
    /**
     * Stops the ripping process.
     * This waits for the Thread to come down
     * and thus blocks
     */     
    public void 
    stop()
    {
        try 
        {
            // Destroy the Process and wait for the Thread 
            // to finish
            _proc.destroy();
            // stop() may be called from _myThread
            // in this case _myThread.join must not be called!
            if (Thread.currentThread()!=_myThread)
                _myThread.join();            
        } catch (Exception ex) 
        {
            ex.printStackTrace();
        }
        
        //TODO find a way to send streamripper a SIG Term for soft shutdown
    }    
        
    /**
     * Parses a Statusmessage received from
     * StreamRipper
     * @param line Line to parse
     */
    private void 
    parseStatusMessage(String line)
    {   
        boolean trackAndTransferredInfoAvailable = false;
        
        // Check if the Process gives us the
        // Stream title
        if (line.startsWith("stream:"))
        {
            // Skip 1 Whitespace
            _name = line.substring("stream:".length() + 1);
        } 
        else if (line.startsWith("Connecting"))
        {
            _status = StreamRipperStatus.CONNECTING;           
        }
        else if (line.startsWith("[buffering"))
        {
            _status = StreamRipperStatus.BUFFERING;
        }
        else if (line.startsWith("[re-connecting"))
       	{
        	_status = StreamRipperStatus.RECONNECTING;
    	}
        else if (line.startsWith("shutting down"))
        {
            _status = StreamRipperStatus.TERMINATED;
        }
        else if (line.startsWith("[ripping"))
        {
            _status = StreamRipperStatus.RIPPING;
            trackAndTransferredInfoAvailable = true;
        }
        else if (line.startsWith("[skipping"))
        {
            _status = StreamRipperStatus.SKIPPING;
            trackAndTransferredInfoAvailable = true;
        }
        else if (line.startsWith("meta interval:")) try
        {
            // check whether meta data is transfered
            // meta interval: -1 is just a hint but does not necessary mean that there is no meta data
            // streamripper won t come beyond Connecting state if no meta data is transfered
            // no files will be saved to hdd in that case, thus end streamripper
            // note: at the moment this issue was only observed on some .ogg streams
            //       perhaps icecast/streamripper bug
            // update: the problem is fixed in streamripper since version 1.64.1
            //         see release notes http://sourceforge.net/project/shownotes.php?release_id=647057&group_id=6172
            //         so in future releases this peace of code should be removed
            //         currently we keep the code to support sr prior version 1.64.1
            // TODO fixed in sr since 1.64.1, remove code when 1.64.1 is minimum required version
            // TODO implement user friendly handling for streams with no metadata, all data saved to -.mp3/ogg is not a good solution

            int metaInterval = Integer.valueOf(line.split(":(\\s)*")[1]);
            if (metaInterval == -1)
            {
                Thread watchdog = new Thread(new Runnable() {
                    public void run()
                    {
                        //wait 15 secs
                        try
                        {
                            Thread.sleep(15000);
                        }
                        catch (InterruptedException e) {}
                        
                        if (StreamRipper.this.getStatus()!=StreamRipperStatus.CONNECTING) return;
                        
                        // if we are still in connecting state, it is likely that there is no meta
                        // data transfered.
                        StreamRipper.this.stop();
                        System.err.println("After 15 seconds still no meta data received for \""
                                            + getStation().getName()
                                            + "\"! Shutting down streamripper.");
                    }
                
                });
                watchdog.start();
                return;
            }
        }
        catch (NumberFormatException nfe ) 
        {
            System.err.println("meta interval parse error: "+nfe.getMessage());
        }

        // update _track and _transfered if we are in Ripping or Skipping state
        if (trackAndTransferredInfoAvailable)
        {
            int trackBegin = line.indexOf("]") + 2;
            int trackEnd = line.lastIndexOf("[") - 1;
            _track = line.substring(trackBegin, trackEnd);
            _transferred = line.substring(trackEnd + 2, line.length()-1).trim();
        }
        
        // Notify observers that we have a new
        // status
        setChanged();
        notifyObservers();
    }

    /**
     * Thread method
     * This does the actual communication across the processes
     */
    public void 
    run()
    {
        try
        {
            // Create Commandline arguments
            // and run Process
            StreamripperCommand srCmd = new StreamripperCommand(Config.getInstance().getProperty(
                    Config.Key.STREAMRIPPER_OPTS));
            srCmd.set(StreamripperCommand.PARAMETER_URL, _station.getUrl());
            srCmd.set(StreamripperCommand.PARAMETER_RELAY_PORT, _relayPort == null ? ""
                    : _relayPort.toString());
            srCmd.set(StreamripperCommand.PARAMETER_DEST_DIR, 
                    Config.getInstance().getProperty(Config.Key.DESTINATION_DIR));
            _proc = srCmd.execute();
            
            // Reader for reading from the Process InputStream
            BufferedReader reader = new BufferedReader(
                    new InputStreamReader(_proc.getInputStream()));
            BufferedReader errors = new BufferedReader(
                    new InputStreamReader(_proc.getErrorStream()));
            
            // Read lines until we should stop running or
            // the process has been finished            
            while (true)
            {                
                String line = reader.readLine();
                
                // If nothing more to read
                // cancel. This means the Process has terminated 
                // and the pipe has broken
                if (line == null)
                    break;
                
                // Else handle next line
                parseStatusMessage(line);
            }
                        
            // Close reader
            reader.close();
            String line;
            while ((line=errors.readLine())!=null)
            {
                if (line.startsWith("error"))
                    _error = line.substring(line.indexOf("[") + 1, line.length() - 1);   
                System.err.println(line);
            }
            errors.close();
            
        } catch (Exception ex)
        {
            // This happens when the process terminates and the pipe 
            // breaks. Could maybe be handled better, but for now
            // the easy solution is not to do anything.
            ex.printStackTrace();
            stop();
        }

        // no more data is received -> TERMINATED
        _status = StreamRipperStatus.TERMINATED;
        setChanged();
        notifyObservers();
    }

    /**
     * Returns the current status
     * @return Status
     */
    public StreamRipperStatus 
    getStatus()
    {
        return _status;
    }

    /**
     * Returns the current track
     * @return currently ripped track
     */
    public String 
    getTrack()
    {
        return _track;
    }

    /**
     * Returns the transferred amount
     * including unit
     * @return amount transferred
     */
    public String 
    getTransferred()
    {
        return _transferred;
    }
     
    /**
     * Returns this Stream's name
     * @return name
     */
    public String 
    getName()
    {
        return _name;
    }
    
    /**
     * Returns the last Error of this Ripper
     */
    public String 
    getError() 
    {
        return _error;
    }
    
    /**
     * Returns this Ripper's Station
     */
    public Station
    getStation() 
    {
        return _station;
    }
   
    public Integer
    getRelayPort()
    {
        return _relayPort;
    }

    public String
    getRelayUrl()
    {
        if (_relayPort==null)
            return null;
        
        String mimeExtension = mimeTypeToExtension.get(getStation().getMimetype());
        if (mimeExtension == null && getStation().getMimetype() != null)
            System.err.println("No mimetype extension defined for \"" + getStation().getMimetype()
                    + "\"");

        return "http://localhost:" + _relayPort
                + ((mimeExtension == null) ? "" : ("/" + mimeExtension));
    }
}
