/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.gui;

import java.awt.CardLayout;
import java.awt.Color;
import java.awt.SystemTray;
import java.awt.Toolkit;
import java.awt.datatransfer.StringSelection;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;

import javax.swing.AbstractAction;
import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.JFrame;
import javax.swing.JLayeredPane;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.LookAndFeel;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.UIManager.LookAndFeelInfo;
import javax.swing.border.LineBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.kuehntopf.streamtastic.core.CompoundMessage;
import org.kuehntopf.streamtastic.core.DirectoryFeed;
import org.kuehntopf.streamtastic.core.DirectoryFeedException;
import org.kuehntopf.streamtastic.core.Message;
import org.kuehntopf.streamtastic.core.RippingManager;
import org.kuehntopf.streamtastic.core.Station;
import org.kuehntopf.streamtastic.core.config.Config;
import org.kuehntopf.streamtastic.core.config.WindowState;
import org.kuehntopf.streamtastic.core.storage.Favorites;
import org.kuehntopf.streamtastic.core.streamripper.StreamripperCommand;
import org.kuehntopf.streamtastic.gui.components.BusyGlassPane;
import org.kuehntopf.streamtastic.gui.components.SidebarMenuCellRenderer;
import org.kuehntopf.streamtastic.gui.views.DirectoryView;
import org.kuehntopf.streamtastic.gui.views.DownloadsView;
import org.kuehntopf.streamtastic.gui.views.FavoritesView;
import org.kuehntopf.streamtastic.gui.views.PreferencesView;
import org.kuehntopf.streamtastic.gui.views.View;
import org.kuehntopf.streamtastic.utils.Player;
import org.kuehntopf.streamtastic.utils.Utilities;

/**
 * Main frame
 * @author  akuehntopf
 */
public class StreamtasticFrame extends javax.swing.JFrame 
{
    public static enum GUIAction {
        CLOSE, ABOUT
    }
    public static enum StationAction {
        PLAY, RECORD, ABORT, ABORT_ALL, ADD_TO_FAVS, REM_FROM_FAVS, COPY_URL_TO_CLIPBOARD
    }
    
    private Map<StationAction, Action> stationActions = null;
    private Map<GUIAction, Action> guiActions = null;
    private StreamtasticTrayIcon trayicon = null;
    private BusyGlassPane busypane = null;
    private AboutFrame aboutFrame;
    private View currentView = null;
    
    /** 
     * Creates new form StreamtasticFrame 
     */
    public 
    StreamtasticFrame() 
    {
        initSharedActions();

        initComponents();
        
        // LAF menu
        initChangeableLookAndFeel();

        // set Frame Icon
        setIconImage(Utilities.loadIcon("streamtasticicon16x16.png").getImage());
                
        // Init Glasspane
        busypane = new BusyGlassPane();
        busypane.setVisible(false);
        switcher.add(busypane, JLayeredPane.MODAL_LAYER);
               
        // Make Directory implementations for shoutcast and icecast available
        for (String feedClassName : new String[] {
                "org.kuehntopf.streamtastic.core.shoutcastfeed.ShoutcastFeed",
                "org.kuehntopf.streamtastic.core.icecastfeed.IcecastFeed" })
            try
            {
                Class.forName(feedClassName);
            }
            catch (Exception e) {}       
        
        // Init Views
        initViews();
        
        // Restore status
        loadWindowStatus();

        // Tray Icon
        if (Config.getInstance().getProperty(Config.Key.MINIMIZE_TO_TRAY).equals("ENABLED")
                && SystemTray.isSupported())
            trayicon = new StreamtasticTrayIcon(this);

        Config.getInstance().addPropertyChangeListener(Config.Key.MINIMIZE_TO_TRAY,
                new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent evt)
                    {
                        if ("ENABLED".equals(evt.getNewValue()))
                        {
                            if (trayicon == null && SystemTray.isSupported())
                                trayicon = new StreamtasticTrayIcon(StreamtasticFrame.this);
                        }
                        else if (trayicon != null)
                        {
                            trayicon.dispose();
                            trayicon = null;
                        }
                    }
                });
    }
    
    private void initSharedActions() 
    {
        stationActions = new HashMap<StationAction, Action>();
        
        stationActions.put(StationAction.PLAY, new AbstractStationAction(
                Utilities.getString("PLAY"), Utilities.getString("PLAY_TOOLTIP"), Utilities
                        .loadIcon("small/play.png")) {

            protected void updateState(Station s)
            {
                if (s==null) setEnabled(false);
                else if (s.isRecording() && s.getRelayUrl()==null) setEnabled(false);
                else setEnabled(true);
            }

            public void actionPerformed(ActionEvent e)
            {
                Station s = (Station) this.getValue(Station.class.getName());
                if (s == null)
                    return;                

                // show Select Player dialog, if player is not available or Ctrl key was pressed
                if ((!Config.getInstance().isValueValidForKey(Config.Key.PLAYER_CMD) || (e
                        .getModifiers() & ActionEvent.CTRL_MASK) != 0))
                {
                    File player = Player.showPlayerSelectionDialog(StreamtasticFrame.this);
                    if (player == null
                            || !Config.Key.PLAYER_CMD.getParameterInfo().isValid(
                                    player.getAbsolutePath()))
                        return;
                    Config.getInstance().setProperty(Config.Key.PLAYER_CMD, player.getAbsolutePath());
                }
                try
                {
                    Player.play(s.isRecording()?s.getRelayUrl():s.getUrl());
                    switcher.repaint();
                }
                catch (Exception ex)
                {
                    Utilities.showError(Utilities.getString("PLAY_ERROR"));
                }
            }
        });
        stationActions.put(StationAction.RECORD, new AbstractStationAction(Utilities
                .getString("RECORD"), Utilities.getString("RECORD"), Utilities
                .loadIcon("small/record.png")) {
            @Override
            protected void updateState(Station s)
            {
                if (s==null) setEnabled(false);
                else if (s.isRecording()) setEnabled(false);
                else setEnabled(true);
            }

            public void actionPerformed(ActionEvent e)
            {
                Station s = (Station) this.getValue(Station.class.getName());
                if (s==null) return;

                // Streamripper available?
                if (!Config.getInstance().isValueValidForKey(Config.Key.STREAMRIPPER_CMD))
                {
                    File location = StreamripperCommand.showSelectStreamripperDialog(StreamtasticFrame.this);
                    if (location==null) return;

                    // Save new Streamripper Command
                    try
                    {
                        Config.getInstance().setProperty(Config.Key.STREAMRIPPER_CMD,
                                location.getAbsolutePath());
                    }
                    catch (IllegalArgumentException ex)
                    {
                        JOptionPane.showMessageDialog(StreamtasticFrame.this, String.format(
                                Utilities.getString("STREAMRIPPER_VERSION_UNSUPPORTED_MSG"),
                                StreamripperCommand.getMinRequiredVersion()), Utilities
                                .getString("STREAMRIPPER_VERSION_UNSUPPORTED_TITLE"),
                                JOptionPane.ERROR_MESSAGE);
                        return;
                    }
                }
                
                // ensure destination directory is set
                if (!Config.getInstance().isValueValidForKey(Config.Key.DESTINATION_DIR))
                {
                    File destDir = Utilities.showFileSelectionDialog(StreamtasticFrame.this,
                            Utilities.getString("SELDEST_TITLE"), Utilities
                                    .getString("SELDEST_APPROVE_LABEL"), Config.getInstance()
                                    .getProperty(Config.Key.DESTINATION_DIR), true);
                    if (destDir == null) return;
                    Config.getInstance().setProperty(Config.Key.DESTINATION_DIR, destDir.getAbsolutePath());
                }

                s.setRecording(true);
                updateStationActions(s);

                switcher.repaint();
            }
        });
        stationActions.put(StationAction.ABORT, new AbstractStationAction(Utilities
                .getString("CANCEL"), Utilities.getString("CANCEL"), Utilities
                .loadIcon("small/cancel.png")) {
            @Override
            protected void updateState(Station s)
            {
                if (s==null) setEnabled(false);
                else if (s.isRecording()) setEnabled(true);
                else setEnabled(false);
            }

            public void actionPerformed(ActionEvent e)
            {
                Station s = (Station) this.getValue(Station.class.getName());
                if (s==null) return;

                s.setRecording(false);
                updateStationActions(s);

                switcher.repaint();
            }
        });
        stationActions.put(StationAction.ABORT_ALL, new AbstractStationAction(Utilities
                .getString("CANCEL_ALL"), Utilities.getString("CANCEL_ALL"), Utilities
                .loadIcon("small/cancel_all.png")) {
            @Override
            protected void updateState(Station s)
            {
                //always active
            }

            public void actionPerformed(ActionEvent e)
            {
                // Ask user for confirmation
                int ret = JOptionPane.showConfirmDialog(StreamtasticFrame.this, 
                        Utilities.getString("DLS_REALLY_CANCEL_ALL"),
                        "Streamtastic",
                        JOptionPane.OK_CANCEL_OPTION);

                // If the user did not press OK: cancel
                if (ret != JOptionPane.OK_OPTION)
                    return;    

                SwingWorker worker = new SwingWorker<Void, Void>() 
                {
                    @Override
                    protected Void doInBackground() throws Exception 
                    {                          
                        RippingManager.getInstance().removeAll();
                        
                        return null;
                    }

                    @Override
                    protected void done() 
                    {   
                        setEnabled(true);
                    }
                 };
                
                 setEnabled(false);
                 worker.execute();
            }
        });
        stationActions.put(StationAction.ADD_TO_FAVS, new AbstractStationAction(Utilities
                .getString("DIR_ADD_TO_FAVS"), Utilities.getString("DIR_ADD_TO_FAVS"), Utilities
                .loadIcon("small/favorites.png")) {
            @Override
            protected void updateState(Station s)
            {
                setEnabled(s!=null && !Favorites.getInstance().isFavorite(s));
            }

            public void actionPerformed(ActionEvent e)
            {
                Station s = (Station) this.getValue(Station.class.getName());
                if (s==null) return;
                
                Favorites.getInstance().addFavorite(s.getUrl(), s.getName(), s.getMimetype(), s.getBitrate());
                updateStationActions(s);
            }
        });
        stationActions.put(StationAction.REM_FROM_FAVS, new AbstractStationAction(Utilities
                .getString("DELETE"), Utilities.getString("DELETE"), Utilities
                .loadIcon("small/delete.png")) {
            @Override
            protected void updateState(Station s)
            {
                setEnabled(s!=null && Favorites.getInstance().isFavorite(s));
            }

            public void actionPerformed(ActionEvent e)
            {
                Station s = (Station) this.getValue(Station.class.getName());
                if (s==null) return;
                
                Favorites.getInstance().removeFavorite(s);
                updateStationActions(s);
            }
        });
        stationActions.put(StationAction.COPY_URL_TO_CLIPBOARD, new AbstractStationAction(Utilities
                .getString("COPY_URL"), Utilities.getString("COPY_URL_TOOLTIP"), Utilities
                .loadIcon("small/copy_url.png")) {
            @Override
            protected void updateState(Station s)
            {
                setEnabled(s!=null);
            }

            public void actionPerformed(ActionEvent e)
            {
                Station s = (Station) this.getValue(Station.class.getName());
                if (s==null) return;
                
                Toolkit.getDefaultToolkit().getSystemClipboard().setContents(new StringSelection(s.getUrl()), null);
            }
        });
        
        //register Observer to be notified about started/stopped rips
        RippingManager.getInstance().addObserver(new Observer() {

            public void update(Observable o, Object msgO)
            {
                CompoundMessage msg = (CompoundMessage) msgO;
                Station s = (Station) msg.getObject();
                if (msg.getMessage() == Message.RIP_REMOVED
                        || msg.getMessage() == Message.RIP_CREATED)
                {
                    if (s.equals(getSharedAction(StationAction.PLAY).getValue(
                            Station.class.getName())))
                        updateStationActions(s);
                    
                    switcher.repaint();
                }
            }

        });

        guiActions = new HashMap<GUIAction, Action>();

        guiActions.put(GUIAction.ABOUT, new AbstractAction(Utilities
                .getString("MAINFRAME_MNU_HELP_ABOUT"), Utilities.loadIcon("small/info.png")) {
            public void actionPerformed(ActionEvent e)
            {
                if (aboutFrame == null)
                {
                    boolean lafDecorated = AboutFrame.isDefaultLookAndFeelDecorated();
                    AboutFrame.setDefaultLookAndFeelDecorated(false);
                    aboutFrame = new AboutFrame();
                    aboutFrame.dispose();
                    aboutFrame.setAlwaysOnTop(true);
                    aboutFrame.setUndecorated(true);
                    ((JPanel) aboutFrame.getContentPane()).setBorder(new LineBorder(Color.GRAY));
                    AboutFrame.setDefaultLookAndFeelDecorated(lafDecorated);
                }
                aboutFrame.setLocationRelativeTo(StreamtasticFrame.this);
                aboutFrame.setVisible(true);
            }
        });
        guiActions.put(GUIAction.CLOSE, new AbstractAction(Utilities
                .getString("MAINFRAME_MNU_FILE_QUIT"), Utilities.loadIcon("small/quit.png")) {
            public void actionPerformed(ActionEvent e)
            {
                dispatchEvent(new WindowEvent(StreamtasticFrame.this, WindowEvent.WINDOW_CLOSING));
            }
        });
    }
    
    public Action getSharedAction(StationAction action)
    {
        return stationActions.get(action);
    }

    public Action getSharedAction(GUIAction action)
    {
        return guiActions.get(action);
    }
    
    public void updateStationActions(Station selectedStation)
    {
        for (Action a:stationActions.values())
            a.putValue(Station.class.getName(), selectedStation);
    }
    
    private void initChangeableLookAndFeel()
    {
        // add look and feel menu
        JMenu lafMenu = new JMenu(Utilities.getString("MAINFRAME_MNU_LAF"));
        ActionListener lafListener = new ActionListener() {
            public void actionPerformed(ActionEvent e)
            {
                JMenuItem mi = (JMenuItem) e.getSource();
                if (mi.getClientProperty("org.kuehntopf.LAFClassName").equals(
                        Config.getInstance().getLookAndFeelClassName()))
                    return;

                Config.getInstance().setProperty(Config.Key.LOOK_AND_FEEL,
                        (String) mi.getClientProperty("org.kuehntopf.LAFClassName"));
            }
        };
        final ButtonGroup lafs = new ButtonGroup();
        String currentLafClassName = UIManager.getLookAndFeel().getClass().getName();
        for (LookAndFeelInfo lafInfo : UIManager.getInstalledLookAndFeels())
        {
            JRadioButtonMenuItem mi = new JRadioButtonMenuItem(lafInfo.getName());
            lafs.add(mi);
            mi.putClientProperty("org.kuehntopf.LAFClassName", lafInfo.getClassName());
            mi.addActionListener(lafListener);
            if (lafInfo.getClassName().equals(currentLafClassName))
                mi.setSelected(true);
            lafMenu.add(mi);
        }
        menubar.add(lafMenu, 1);

        // apply changed look and feel 
        Config.getInstance().addPropertyChangeListener(Config.Key.LOOK_AND_FEEL, new PropertyChangeListener() {
            private final String RESTART_NEEDED = " ("
                + Utilities.getString("MAINFRAME_MNU_LAF_RESTART_NEEDED") + ")";
            private AbstractButton prevItem = null;
            private LookAndFeel currentLaf, newLaf;

            private void checkRestartNeeded(AbstractButton mi)
            {
                // clear restart needed
                if (prevItem != null) {
                    prevItem.setText(prevItem.getText().substring(0,
                            prevItem.getText().length() - RESTART_NEEDED.length()));
                    prevItem = null;
                }

                if (UIManager.getLookAndFeel() == newLaf
                                && newLaf.getSupportsWindowDecorations() == JFrame
                                        .isDefaultLookAndFeelDecorated())
                            return; //everything is fine 

                if (UIManager.getLookAndFeel() != newLaf)
                    JOptionPane.showMessageDialog(StreamtasticFrame.this, Utilities
                            .getString("MAINFRAME_MNU_LAF_RESTART_NEEDED_MSG"), Utilities
                            .getString("MAINFRAME_MNU_LAF_RESTART_NEEDED"),
                            JOptionPane.INFORMATION_MESSAGE);
                else 
                    JOptionPane.showMessageDialog(StreamtasticFrame.this, Utilities
                            .getString("MAINFRAME_MNU_LAF_RESTART_RECMDED_MSG"), Utilities
                            .getString("MAINFRAME_MNU_LAF_RESTART_RECMDED"),
                            JOptionPane.INFORMATION_MESSAGE);

                // set restart needed label
                prevItem = mi;
                if (prevItem!=null)
                    prevItem.setText(prevItem.getText() + RESTART_NEEDED);
            }

            public void propertyChange(PropertyChangeEvent evt)
            {
                currentLaf = UIManager.getLookAndFeel();
                try
                {
                    newLaf = (LookAndFeel) Class.forName(
                                    Config.getInstance().getLookAndFeelClassName()).newInstance();
                    if (currentLaf.equals(newLaf))
                        return;

                    // there will be no window decorations with activated default LookAndFeel decorations 
                    // in case the new design does not supply some
                    // thanks to J. Ulbts for his hints on this
                    // for details, see http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=5050922
                    if (!newLaf.getSupportsWindowDecorations()
                            && JFrame.isDefaultLookAndFeelDecorated())
                        System.err.println("default LAF window decorations activated "
                                + "but not supported by current LAF!");
                    else 
                    {
                        UIManager.setLookAndFeel(newLaf);
                        SwingUtilities.updateComponentTreeUI(StreamtasticFrame.this);
                        if (aboutFrame!=null) SwingUtilities.updateComponentTreeUI(aboutFrame);
                        StreamtasticFrame.this.pack();
                    }
                }
                catch (Exception e1)
                {
                    // This happens with current used substance version
                    // switching to substance works fine but it is not possible to successfully 
                    // switch from substance to another laf
                    // uninitialize of substance seems to be buggy
                    //   -> exception thrown and gui is left in an unusable state
                    // TODO reevaluate with new substance versions
                    System.err.println("Error setting LAF, trying to restore old LAF!\n\tcause: "
                            + e1.getMessage());
                    // fallback to old laf
                    try
                    {
                        UIManager.setLookAndFeel(currentLaf);
                        SwingUtilities.updateComponentTreeUI(StreamtasticFrame.this);
                        StreamtasticFrame.this.pack();
                    }
                    catch (UnsupportedLookAndFeelException e2)
                    {
                        // should never happen
                        e2.printStackTrace();
                    }
                }

                // synchronize selection in design menu if needed
                String lafclassName = Config.getInstance().getLookAndFeelClassName();
                Enumeration<AbstractButton> mis = lafs.getElements();
                AbstractButton mi = null;
                while (mis.hasMoreElements()) 
                {
                    // check client property, adjust selection
                    mi = mis.nextElement();
                    if (mi.getClientProperty("org.kuehntopf.LAFClassName").equals(lafclassName))
                    {
                        if (!mi.isSelected()) 
                            lafs.setSelected(mi.getModel(), true);
                        break;
                    }
                }
                if (mi==null) lafs.clearSelection();
                checkRestartNeeded(mi);
            }
        });
    }

    /**
     * Initializes the Sidebar
     */
    private void 
    initViews()
    {
        final CardLayout cardLayout = new CardLayout();
        switcher.setLayout(cardLayout);
        
         // Setup list
        sidebarmenu.setCellRenderer(new SidebarMenuCellRenderer(sidebarmenu));
        sidebarmenu.setFixedCellHeight(30);
        
        // init model
        DefaultListModel sidebarmodel= new DefaultListModel();
        sidebarmodel.addElement(new FavoritesView(this));
        for (String url : DirectoryFeed.getURLsForRegisteredFeeds())
            try
            {
                sidebarmodel
                        .addElement(new DirectoryView(this, DirectoryFeed.getDirectoryFeed(url)));
            }
            catch (DirectoryFeedException e)
            {
                e.printStackTrace();
            }
        sidebarmodel.addElement(new DownloadsView(this));
        sidebarmodel.addElement(new PreferencesView(this));
        sidebarmenu.setModel(sidebarmodel);

        Enumeration<View> views = (Enumeration<View>) sidebarmodel.elements();
        while (views.hasMoreElements()) 
        {
            View view = views.nextElement();
            switcher.add(view, view.getDisplayName());
        }
        
        // Add listener for reacting to selection changes
        sidebarmenu.addListSelectionListener(new ListSelectionListener() 
        {
            public void valueChanged(ListSelectionEvent evt) 
            {
                if (evt.getValueIsAdjusting() == true)
                    return;
                
                View view = (View) sidebarmenu.getSelectedValue();
                if (currentView!=null && currentView!=view && !currentView.isReadyForViewChange())
                    sidebarmenu.setSelectedValue(currentView, true);
                else if (currentView!=view)
                {
                    busypane.setVisible(false);
                    cardLayout.show(switcher, view.getDisplayName());
                    currentView = view;
                }
            }
        });
        
        // this ensures that the first view is activated properly
        ((View)sidebarmenu.getModel().getElementAt(0)).setVisible(false);
        // Select item 0 by default
        sidebarmenu.setSelectedIndex(0);
    }
    
    /**
     * Sets the busy status
     * @param busy
     */
    public void 
    setBusyPaneVisibleForView(View view, boolean busy)            
    {
        if (sidebarmenu.getSelectedValue()!=null && !sidebarmenu.getSelectedValue().equals(view))
            return;
        busypane.setVisible(busy);
    }
    
    /**
     * Loads the current Window location
     * from the Config
     */
    private void
    loadWindowStatus()
    {
        WindowState wState = Config.getInstance().getWindowState();
        if (wState == null)
        {
            setLocationByPlatform(true);
            wState = new WindowState(getLocation() ,getSize(), getExtendedState());
            Config.getInstance().setProperty(Config.Key.WINDOW_STATE, wState.toString());
            return;
        }

        setLocation(wState.getLocation());
        setSize(wState.getSize());
        setExtendedState(wState.getState());
    }
    
    /**
     * Saves the current Window location to the Config
     */
    private void
    saveWindowStatus()
    {
        WindowState wState = Config.getInstance().getWindowState();
        if (wState == null)
            return;

        // Only save bounds if the Window state is normal
        if (getExtendedState() == JFrame.NORMAL)
        {
            wState.setLocation(getLocation());
            wState.setSize(getSize());
        }

        // don't save minimized window state
        if ((getExtendedState() & JFrame.ICONIFIED) == 0)
            wState.setState(getExtendedState());

        Config.getInstance().setProperty(Config.Key.WINDOW_STATE, wState.toString());
    }

    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        splitpane = new javax.swing.JSplitPane();
        scrollpane = new javax.swing.JScrollPane();
        sidebarmenu = new javax.swing.JList();
        switcher = new javax.swing.JLayeredPane();
        menubar = new javax.swing.JMenuBar();
        mnuFile = new javax.swing.JMenu();
        mnuHelp = new javax.swing.JMenu();

        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        setTitle(Utilities.getString("MAINFRAME_TITLE")); // NOI18N
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                formWindowClosing(evt);
            }
        });

        splitpane.setDividerLocation(150);
        splitpane.setDividerSize(4);

        scrollpane.setPreferredSize(new java.awt.Dimension(150, 430));

        sidebarmenu.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        scrollpane.setViewportView(sidebarmenu);

        splitpane.setLeftComponent(scrollpane);

        switcher.setPreferredSize(new java.awt.Dimension(690, 430));
        splitpane.setRightComponent(switcher);

        getContentPane().add(splitpane, java.awt.BorderLayout.CENTER);

        mnuFile.setText(Utilities.getString("MAINFRAME_MNU_FILE")); // NOI18N
        mnuFile.add(getSharedAction(GUIAction.CLOSE));
        menubar.add(mnuFile);

        mnuHelp.setText(Utilities.getString("MAINFRAME_MNU_HELP")); // NOI18N
        mnuHelp.add(getSharedAction(GUIAction.ABOUT));
        menubar.add(mnuHelp);

        setJMenuBar(menubar);

        pack();
    }// </editor-fold>//GEN-END:initComponents

    /**
     * Reacts to X button presses
     * @param evt
     */
    private void formWindowClosing(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowClosing
        //in case there are running rips, ask user if they should be aborted
        if (RippingManager.getInstance().getRippedStations().length > 0
            && JOptionPane.showOptionDialog(
                this, Utilities.getString("MAINFRAME_DLG_WARNING_RUNNING_RIPS_ON_APP_CLOSE_MSG"),
                Utilities.getString("MAINFRAME_DLG_WARNING_RUNNING_RIPS_ON_APP_CLOSE_TITLE"),
                JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE, null,
                new Object[] {
                    Utilities.getString("MAINFRAME_DLG_WARNING_RUNNING_RIPS_ON_APP_CLOSE_OPT_ABORT_RIPS"),
                    Utilities.getString("MAINFRAME_DLG_WARNING_RUNNING_RIPS_ON_APP_CLOSE_OPT_CANCEL_CLOSE") 
                },
                Utilities.getString("MAINFRAME_DLG_WARNING_RUNNING_RIPS_ON_APP_CLOSE_OPT_CANCEL_CLOSE")) == 1)
            return;

        //cancel all downloads and save window state when streamtastic is closed
        saveWindowStatus();
        RippingManager.getInstance().removeAll();
        // remove trayicon, otherwise streamtastic won't stop if closed from trayicon popup menu
        if (trayicon != null)
            trayicon.dispose();
        // dispose about dialog
        if (aboutFrame != null) aboutFrame.dispose();
        // dispose main frame
        this.dispose();
    }//GEN-LAST:event_formWindowClosing

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenuBar menubar;
    private javax.swing.JMenu mnuFile;
    private javax.swing.JMenu mnuHelp;
    private javax.swing.JScrollPane scrollpane;
    private javax.swing.JList sidebarmenu;
    private javax.swing.JSplitPane splitpane;
    private javax.swing.JLayeredPane switcher;
    // End of variables declaration//GEN-END:variables
    
    private abstract class AbstractStationAction extends AbstractAction {

        public AbstractStationAction(String name, String tooltip, Icon icon)
        {
            super(name, icon);
            putValue(Action.SHORT_DESCRIPTION, tooltip);
        }

        @Override
        public void putValue(String key, Object newValue)
        {
            super.putValue(key, newValue);
            
            if (!key.equals(Station.class.getName())) return;
            
            updateState((Station) newValue);
        }
        
        protected abstract void updateState(Station s);
    }
}
