/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.gui.components;

import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.lang.reflect.Method;
import java.util.Comparator;
import java.util.EventObject;

import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.RowSorter;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.event.CellEditorListener;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import org.kuehntopf.streamtastic.core.Station;
import org.kuehntopf.streamtastic.utils.Player;
import org.kuehntopf.streamtastic.utils.Utilities;

/**
 * Column Class for Station objects. Makes creating table models based on Station easier.
 * 
 * @author mikeL@launchpad.net
 */
public class StationColumn extends TableColumn {

    /**
     * 
     */
    private static final long serialVersionUID = 1L;
    
    private static ImageIcon record = Utilities.loadIcon("small/record.png");
    private static ImageIcon play = Utilities.loadIcon("small/play.png");
    private static ImageIcon record_play = initRecordPlayIcon();
    private static Border border = null;

    private Method getMethod;
    private Method setMethod;
    private boolean showRecordingIcon;
    
    private static ImageIcon initRecordPlayIcon() 
    {
        int width = Math.max(record.getIconWidth(), play.getIconWidth());
        int height = Math.max(record.getIconHeight(), play.getIconHeight());
        
        BufferedImage img = new BufferedImage(width,height,BufferedImage.TYPE_INT_ARGB);
        Graphics2D g2d = (Graphics2D) img.getGraphics();
        
        g2d.drawImage(record.getImage(), (width-record.getIconWidth())/2, (height-record.getIconHeight())/2, null);
        g2d.drawImage(play.getImage(), (width-play.getIconWidth())/2, (height-play.getIconHeight())/2, null);
        
        return new ImageIcon(img);
    }

    public StationColumn(Station.Property type, int width, boolean showRecordingIcon)
    {
        super(type.ordinal(),width);
        String title;
        switch (type) {
            case NAME: title = Utilities.getString("DIR_COL_NAME"); break;
            case CURRENT_TRACK: title = Utilities.getString("DIR_COL_CURRENT"); break;
            case LISTENERS: title = Utilities.getString("DIR_COL_LISTENERS"); break;
            case MIMETYPE: title = Utilities.getString("DIR_COL_MIMETYPE"); break;
            case BITRATE: title = Utilities.getString("DIR_COL_BITRATE"); break;
            case GENRE: title = Utilities.getString("DIR_COL_GENRE"); break;
            case URL: title = Utilities.getString("FAVS_COL_URL"); break;
            case RECORDING_STATE: title = Utilities.getString("DLS_COL_STATUS"); break;
            case RECORDED_DATA_AMOUNT: title = Utilities.getString("DLS_COL_TRANSFERRED"); break;
            default: title = null;
        }
        init(type, title, showRecordingIcon);
    }

    public StationColumn(Station.Property type,String title, int width, boolean showRecordingIcon)
    {
        super(type.ordinal(),width);
        init(type,title,showRecordingIcon);
    }

    public StationColumn(Station.Property type, int width)
    {
        this(type, width ,true);
    }

    public StationColumn(Station.Property type)
    {
        this(type, (type==Station.Property.CURRENT_TRACK||type==Station.Property.NAME)?150:30, true);
    }

    private void init(final Station.Property type, String title, boolean showRecordingIcon)
    {
        this.setIdentifier(type);
        this.showRecordingIcon = showRecordingIcon;

        String setMethodName = "set" + type.toString();

        getMethod = getGetterMethod(type);
        for (Method m : Station.class.getMethods())
            if (m.getName().equalsIgnoreCase(setMethodName) && m.getParameterTypes().length == 1
                    && getMethod.getReturnType() == m.getParameterTypes()[0])
            {
                setMethod = m;
                break;
            }

        setHeaderValue(title);

        setCellRenderer(new TableCellRenderer() {
            @Override
            public Component getTableCellRendererComponent(JTable table, Object value,
                    boolean isSelected, boolean hasFocus, int row, int column)
            {

                Station s = (Station) value;
                if (getMethod == null)
                    value = "no get-method for \"" + type + "\" in class Station";
                try
                {
                    value = getMethod.invoke(s);
                } catch (Exception e)
                {
                    value = "Error invoking get-method for \"" + type + "\" in class Station";
                    e.printStackTrace();
                }

                Component c = table.getDefaultRenderer(
                        value != null ? value.getClass() : Object.class)
                        .getTableCellRendererComponent(table, value, isSelected, hasFocus, row,
                                column);

                if (c instanceof JLabel) 
                {
                    JLabel l = (JLabel) c;
                    // disable cell selection highlighting
                    if (border==null && l.getBorder() instanceof EmptyBorder) border = l.getBorder();
                    l.setBorder(border);

                    // add record or play icon
                    l.setIcon(null);
                    //show the icon in the first column only
                    if (column==0) {
                        if (StationColumn.this.showRecordingIcon && s.isRecording()) 
                            l.setIcon(record);
                        if (Player.isPlaying(s.getUrl()) || Player.isPlaying(s.getRelayUrl())) 
                            l.setIcon(l.getIcon()==null?play:record_play);
                    } 
                    // only show tooltip when cell contents is truncated
                    if (l.getPreferredSize().getWidth() > getWidth())
                        l.setToolTipText(l.getText());
                    else l.setToolTipText(null);
                }

                return c;
            }

        });

        setCellEditor(new TableCellEditor() {

            private final TableCellEditor cellEditor = getMethod == null ? null : new JTable()
                    .getDefaultEditor(getMethod.getReturnType());
            private Object target;

            @Override
            public boolean stopCellEditing()
            {
                return cellEditor.stopCellEditing();
            }

            @Override
            public boolean shouldSelectCell(EventObject anEvent)
            {
                return cellEditor.shouldSelectCell(anEvent);
            }

            @Override
            public void removeCellEditorListener(CellEditorListener l)
            {
                cellEditor.removeCellEditorListener(l);
            }

            @Override
            public boolean isCellEditable(EventObject anEvent)
            {
                // do not start editing on mouse clicks
                if (anEvent instanceof MouseEvent) return false;
                return cellEditor.isCellEditable(anEvent);
            }

            @Override
            public Object getCellEditorValue()
            {
                try
                {
                    setMethod.invoke(target, cellEditor.getCellEditorValue());
                } catch (Exception e)
                {
                    e.printStackTrace();
                }
                return cellEditor.getCellEditorValue();
            }

            @Override
            public void cancelCellEditing()
            {
                cellEditor.cancelCellEditing();
            }

            @Override
            public void addCellEditorListener(CellEditorListener l)
            {
                cellEditor.addCellEditorListener(l);
            }

            @Override
            public Component getTableCellEditorComponent(JTable table, Object value,
                    boolean isSelected, int row, int column)
            {
                target = value;
                try
                {
                    value = getMethod.invoke((Station) value);
                } catch (Exception e)
                {
                    value = "Error invoking get-method for \"" + type + "\" in class Station";
                    e.printStackTrace();
                }
                return cellEditor
                        .getTableCellEditorComponent(table, value, isSelected, row, column);
            }
        });
    }

    static Method getGetterMethod(Station.Property type)
    {
        for (Method m : Station.class.getMethods())
            if (m.getName().equalsIgnoreCase("get" + type) && m.getParameterTypes().length == 0)
            {
                return m;
            }
        return null;
    }

    public static RowSorter<TableModel> getRowsorter(TableModel model)
    {
        TableRowSorter<TableModel> rowsorter = new TableRowSorter<TableModel>(model);
        for (Station.Property p : Station.Property.values())
            rowsorter.setComparator(p.ordinal(), new StationComparator(p));

        return rowsorter;
    }

    private static class StationComparator implements Comparator<Station> {

        private Method getMethod;

        public StationComparator(Station.Property type)
        {
            getMethod = getGetterMethod(type);
        }

        @Override
        public int compare(Station o1, Station o2)
        {
            try
            {
                Object v1 = getMethod.invoke(o1);
                Object v2 = getMethod.invoke(o2);

                //handle null values
                if (v1 == null)
                    return v2 == null ? 0 : -1;
                else if (v2 == null)
                    return 1;

                // use Collator for Strings
                // otherwise "RAT" is listed before "rad"
                if (v1 instanceof String)
                    return java.text.Collator.getInstance().compare(v1, v2);

                // compare using Comparable interface
                // works for all clases implementing Comparable
                // for instance String, Integer, Boolean etc
                if (v1 instanceof Comparable)
                {
                    return ((Comparable) v1).compareTo(v2);
                }
            } catch (Exception e)
            {
                e.printStackTrace();
            }

            return 0;
        }
    }
}