/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.gui.models;

import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import javax.swing.Box;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SpinnerNumberModel;

import org.kuehntopf.streamtastic.core.config.Config;
import org.kuehntopf.streamtastic.core.config.Config.Key;
import org.kuehntopf.streamtastic.core.streamripper.StreamripperCommand;
import org.kuehntopf.streamtastic.core.streamripper.StreamripperCommand.Option;
import org.kuehntopf.streamtastic.gui.components.CustomEditor;
import org.kuehntopf.streamtastic.gui.components.ExpandableOptionPane;
import org.kuehntopf.streamtastic.gui.components.ExpandableOptionPane.ParamType;
import org.kuehntopf.streamtastic.utils.Utilities;

/**
 * @author Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 *
 */
public class StreamRipperOptionModel extends OptionModel {

    private static enum OverwriteMode {always, never, larger, version};
    private static enum Id3 {v1, v2};

    private Map<Option, String> defaultOptions;

    private List<ExpandableOptionPane> optionPanes;
    private ExpandableOptionPane outPutOptions, id3tags, conSettings;
    private AdvancedPreferencesModel model;
    private JScrollPane pane;

    private Key[] configParams = new Key[] {
            Key.STREAMRIPPER_OPTS,
            Key.STREAMRIPPER_RELAY_PORT_RANGE
    };

    public StreamRipperOptionModel(AdvancedPreferencesModel model)
    {
        this.model = model;

        // init defaults
        try
        {
            defaultOptions = StreamripperCommand.decompileOptionString(Config.getInstance()
                    .getDefaultValue(Key.STREAMRIPPER_OPTS));
            setStreamRipperOptions(defaultOptions);
            setRelayPortRange(Config.getInstance().getDefaultValue(Key.STREAMRIPPER_RELAY_PORT_RANGE));
        }
        catch (RuntimeException e)
        {
            e.printStackTrace();
        }
    }

    private JScrollPane initDisplay()
    {
        optionPanes = new ArrayList<ExpandableOptionPane>();

        // constraints
        GridBagConstraints optionPaneConstraints = new GridBagConstraints();
        optionPaneConstraints.fill = GridBagConstraints.HORIZONTAL;
        optionPaneConstraints.gridx = GridBagConstraints.REMAINDER;
        optionPaneConstraints.insets = new Insets(5,5,5,5);
        optionPaneConstraints.weightx = 1;
        optionPaneConstraints.anchor = GridBagConstraints.NORTH;

        GridBagConstraints spacerConstraints = (GridBagConstraints) optionPaneConstraints.clone();
        spacerConstraints.fill = GridBagConstraints.BOTH;
        spacerConstraints.weighty = 1;

        // TODO Provide tooltips

        // connection settings
        conSettings = new ExpandableOptionPane(Utilities.getString("PREF_TAB_SR_EXPANE_CONNECTIONS_SETTINGS_TITLE"));
        conSettings.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_CONNECTION_SETTINGS_TITLE"));
        conSettings.addParam(Utilities.getString("PREF_TAB_SR_PARAM_PROXY_LABEL"), Option.PROXY, ParamType.CB_LONG_STRING);
        conSettings.addParam(Utilities.getString("PREF_TAB_SR_PARAM_USERAGENT_LABEL"), Option.USER_AGENT, ParamType.CB_LONG_STRING);
        conSettings.addParam(Utilities.getString("PREF_TAB_SR_PARAM_DISABLE_RECONNECT_LABEL"), Option.DISABLE_RECONNECT, ParamType.CB_SIMPLE); 
        conSettings.addParam(Utilities.getString("PREF_TAB_SR_PARAM_STREAM_STUTTER_TIMEOUT_LABEL"), Option.STREAM_STUTTER_TIMEOUT, ParamType.CB_SPINNER);
        {
            conSettings.setModelForSpinner(Option.STREAM_STUTTER_TIMEOUT, new SpinnerNumberModel(30,0,Integer.MAX_VALUE,1));
        }
        conSettings.addParam(Utilities.getString("PREF_TAB_SR_PARAM_CREATE_RELAY_STREAM_LABEL"), Key.STREAMRIPPER_RELAY_PORT_RANGE, ParamType.CB_SHORT_STRING);
        conSettings.addParam(Utilities.getString("PREF_TAB_SR_PARAM_MAX_RELAY_STREAM_CONNECTIONS_LABEL"), Option.MAX_RELAY_STREAM_CONNECTIONS, ParamType.CB_SPINNER);
        {
            conSettings.setModelForSpinner(Option.MAX_RELAY_STREAM_CONNECTIONS, new SpinnerNumberModel(1,1,Integer.MAX_VALUE,1));
        }
        optionPanes.add(conSettings);
        
        // output options
        outPutOptions = new ExpandableOptionPane(Utilities.getString("PREF_TAB_SR_EXPANE_OUTPUT_OPTIONS_TITLE"));
        outPutOptions.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_OUTPUT_OPTIONS_TITLE"));
        //TODO readd destination dir parameter on introduction of ripping profiles
        //outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_DESTINATION_DIRECTORY_LABEL"), Key.DESTINATION_DIR, ParamType.CB_LONG_STRING);
        outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_NO_SEPARATE_DIRECTORY_PER_STATION_LABEL"), Option.NO_SEPARATE_DIRECTORY_PER_STATION, ParamType.CB_SIMPLE);
        outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_RIP_TO_SINGLE_FILE_LABEL"), Option.RIP_TO_SINGLE_FILE, ParamType.CB_SHORT_STRING);
        {
            outPutOptions.setTooltipForComponents(Option.RIP_TO_SINGLE_FILE, null, Utilities.getString("PREF_TAB_SR_PARAM_RIP_TO_SINGLE_FILE_TXTFIELD_TOOLTIP"));
        }
        outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_DONT_CREATE_TRACKS_LABEL"), Option.DONT_CREATE_TRACKS, ParamType.CB_SIMPLE);
        outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_SKIP_TRACKS_ON_START_LABEL"), Option.SKIP_TRACKS_ON_START, ParamType.CB_SPINNER);
        {
            outPutOptions.setModelForSpinner(Option.SKIP_TRACKS_ON_START, new SpinnerNumberModel(0,0,Integer.MAX_VALUE,1));
        }
        outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_TRACK_PADDING_LABEL"), Option.TRACK_PADDING, ParamType.CB_2SPINNER);
        {
            outPutOptions.setModelForSpinner(Option.TRACK_PADDING, new SpinnerNumberModel(0,0,Integer.MAX_VALUE,1), new SpinnerNumberModel(0,0,Integer.MAX_VALUE,1));
        }
        outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_ADD_SEQ_NUMBER_TO_TRACKS_LABEL"), Option.ADD_SEQ_NUMBER_TO_TRACKS, ParamType.CB_SPINNER); 
        {
            outPutOptions.setModelForSpinner(Option.ADD_SEQ_NUMBER_TO_TRACKS, new SpinnerNumberModel(1,1,Integer.MAX_VALUE,1));
        }
        outPutOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_FILENAME_PATTERN_LABEL"), Option.FILENAME_PATTERN, ParamType.CB_SHORT_STRING);
        {
            outPutOptions.setTooltipForComponents(Option.FILENAME_PATTERN, null, Utilities.getString("PREF_TAB_SR_PARAM_FILENAME_PATTERN_TXTFIELD_TOOLTIP"));
        }
        optionPanes.add(outPutOptions);

        // duplicates handling
        ExpandableOptionPane duplicatesOptions = new ExpandableOptionPane(Utilities.getString("PREF_TAB_SR_EXPANE_DUPLCATES_HANDLING_TITLE"));
        duplicatesOptions.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_DUPL_COMPL_TRACKS_TITLE"));
        duplicatesOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_OVERWRITE_ALWAYS_LABEL"), OverwriteMode.always, ParamType.RB_SIMPLE);
        duplicatesOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_OVERWRITE_NEVER_LABEL"), OverwriteMode.never, ParamType.RB_SIMPLE);
        duplicatesOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_OVERWRITE_LARGER_LABEL"), OverwriteMode.larger, ParamType.RB_SIMPLE);
        duplicatesOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_OVERWRITE_VERSION_LABEL"), OverwriteMode.version, ParamType.RB_SIMPLE);
        duplicatesOptions.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_DUPL_INCOMPL_TRACKS_TITLE"));
        duplicatesOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_DONT_OVERWRITE_TRACKS_IN_INCOMPLETE_LABEL"), Option.DONT_OVERWRITE_TRACKS_IN_INCOMPLETE, ParamType.CB_SIMPLE);
        duplicatesOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_TRUNCATE_TRACKS_IN_INCOMPLETE_LABEL"), Option.TRUNCATE_TRACKS_IN_INCOMPLETE, ParamType.CB_SIMPLE);
        optionPanes.add(duplicatesOptions);

        // stop conditions
        ExpandableOptionPane stopConditions = new ExpandableOptionPane(Utilities.getString("PREF_TAB_SR_EXPANE_STOP_CONDITIONS_TITLE"));
        stopConditions.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_STOP_CONDITIONS_TITLE"));
        stopConditions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_TIME_LIMIT_LABEL"), Option.TIME_LIMIT, ParamType.CB_SPINNER);
        {
            stopConditions.setModelForSpinner(Option.TIME_LIMIT, new SpinnerNumberModel(1,1,Integer.MAX_VALUE,1));
        }
        stopConditions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_SPACE_LIMIT_LABEL"), Option.SPACE_LIMIT, ParamType.CB_SPINNER);
        {
            stopConditions.setModelForSpinner(Option.SPACE_LIMIT, new SpinnerNumberModel(1,1,Integer.MAX_VALUE,1));
        }
        optionPanes.add(stopConditions);

        // id3tags options
        id3tags = new ExpandableOptionPane(Utilities.getString("PREF_TAB_SR_EXPANE_ID3TAG_OPTIONS_TITLE"));
        id3tags.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_ID3TAG_OPTIONS_TITLE"));
        id3tags.addParam(Utilities.getString("PREF_TAB_SR_PARAM_ADD_ID3V1_LABEL"), Id3.v1, ParamType.CB_SIMPLE); 
        id3tags.addParam(Utilities.getString("PREF_TAB_SR_PARAM_ADD_ID3V2_LABEL"), Id3.v2, ParamType.CB_SIMPLE); 
        optionPanes.add(id3tags);

        // silence detection
        ExpandableOptionPane silenceOptions = new ExpandableOptionPane(Utilities.getString("PREF_TAB_SR_EXPANE_SILENCE_DETECTION_TITLE"));
        silenceOptions.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_SILENCE_DETECTION_TITLE"));
        silenceOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_NO_SILENCE_DETECTION_LABEL"), Option.NO_SILENCE_DETECTION, ParamType.CB_SIMPLE);
        silenceOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_SILENCE_DETECTION_OFFSET_LABEL"), Option.SILENCE_DETECTION_OFFSET_ON_TRACK_CHANGE, ParamType.CB_SPINNER);
        {
            silenceOptions.setModelForSpinner(Option.SILENCE_DETECTION_OFFSET_ON_TRACK_CHANGE, new SpinnerNumberModel(0,Integer.MIN_VALUE,Integer.MAX_VALUE,1));
        }
        silenceOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_SILENCE_SEARCH_WINDOW_LABEL"), Option.SILENCE_SEARCH_WINDOW, ParamType.CB_2SPINNER);
        {
            silenceOptions.setModelForSpinner(Option.SILENCE_SEARCH_WINDOW, new SpinnerNumberModel(1,1,Integer.MAX_VALUE,1), new SpinnerNumberModel(1,1,Integer.MAX_VALUE,1));
        }
        silenceOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_MIN_SILENCE_LENGTH_LABEL"), Option.MIN_SILENCE_LENGTH_FOR_SPLIT, ParamType.CB_SPINNER);
        {
            silenceOptions.setModelForSpinner(Option.MIN_SILENCE_LENGTH_FOR_SPLIT, new SpinnerNumberModel(1,1,Integer.MAX_VALUE,1));
        }
        silenceOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_ALT_SILENCE_DETECTION_LABEL"), Option.ALTERNATIVE_SILENCE_DETECTION, ParamType.CB_SIMPLE);
        optionPanes.add(silenceOptions);

        // metaData options
        ExpandableOptionPane metaDataOptions = new ExpandableOptionPane(Utilities.getString("PREF_TAB_SR_EXPANE_METADATA_OPTIONS_TITLE"));
        metaDataOptions.startGroup(Utilities.getString("PREF_TAB_SR_GROUP_METADATA_OPTIONS_TITLE"));
        metaDataOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_META_DATA_PARSING_RULES_LABEL"), Option.META_DATA_PARSING_RULES, ParamType.CB_LONG_STRING);
        {
            metaDataOptions.setCustomEditorForTextField(Option.META_DATA_PARSING_RULES, new CustomEditor() {
                public String edit(JComponent parent, String currentValue)
                {
                    File file = Utilities.showFileSelectionDialog(
                        parent, Utilities.getString("PREF_TAB_SR_PARAM_META_DATA_PARSING_RULES_DLG_TITLE"),
                        Utilities.getString("PREF_TAB_SR_PARAM_META_DATA_PARSING_RULES_DLG_APPROVE_LABEL"),
                        currentValue, false
                    );
                    return (file==null)? currentValue: file.getAbsolutePath();
                }
            });
        }
        metaDataOptions.addParam(Utilities.getString("PREF_TAB_SR_PARAM_META_DATA_PROVIDER_LABEL"), Option.META_DATA_PROVIDER, ParamType.CB_LONG_STRING); 
        {
            metaDataOptions.setCustomEditorForTextField(Option.META_DATA_PROVIDER, new CustomEditor() {
                public String edit(JComponent parent, String currentValue)
                {
                    File file = Utilities.showFileSelectionDialog(
                        parent, Utilities.getString("PREF_TAB_SR_PARAM_META_DATA_PROVIDER_DLG_TITLE"),
                        Utilities.getString("PREF_TAB_SR_PARAM_META_DATA_PROVIDER_DLG_APPROVE_LABEL"),
                        currentValue, false
                    );
                    return (file==null)? currentValue: file.getAbsolutePath();
                }
            });
        }
        optionPanes.add(metaDataOptions);

        // main panel
        JPanel main = new JPanel();
        main.setLayout(new GridBagLayout());
        for (ExpandableOptionPane oPane : optionPanes)
            main.add(oPane, optionPaneConstraints);
        //add spacer
        main.add(Box.createVerticalGlue(),spacerConstraints);

        JScrollPane display = new JScrollPane(main);
        display.getVerticalScrollBar().setUnitIncrement(20);
        return display;
    }

    public Container getDisplay() 
    {
        if (pane==null) pane = initDisplay();
        return pane;
    }

    private void setRelayPortRange(String value)
    {
        conSettings.setValue(Key.STREAMRIPPER_RELAY_PORT_RANGE, value!=null&&value.length()==0?null:value);
        if (model.isModified(Key.STREAMRIPPER_RELAY_PORT_RANGE)
                || !Config.getInstance().isDefaultValue(Key.STREAMRIPPER_RELAY_PORT_RANGE))
            conSettings.setExpanded(true);
    }

    private String getRelayPortRange() 
    {
        if (!conSettings.isExpanded()) return Config.getInstance().getDefaultValue(Key.STREAMRIPPER_RELAY_PORT_RANGE);
        return conSettings.getValue(Key.STREAMRIPPER_RELAY_PORT_RANGE);
    }

    private void setStreamRipperOptions(Map<Option, String> options)
    {
        for (ExpandableOptionPane oPane : optionPanes)
        {
            boolean isDefault = defaultOptions != null;

            for (Enum param : oPane.getParams())
            {
                if (param instanceof Option)
                {
                    String value = options.get((Option)param);
                    value = value == null && options.containsKey(param) ? "" : value;

                    // switch values for Option.TRACK_PADDING
                    if (param == Option.TRACK_PADDING && value!=null && value.contains(":"))
                    {
                        String[] splitted = value.split(":");
                        value = splitted[1]+":"+splitted[0];
                    }

                    oPane.setValue(param, value);
                }
                else if (param instanceof OverwriteMode)
                    oPane.setValue(param, param.toString().equals(
                            options.get(Option.OVERWRITE_COMPLETE_TRACKS)) ? "" : null);
                else if (param == Id3.v1)
                    oPane.setValue(param, (options.containsKey(Option.ADD_ID3V1) && !options
                            .containsKey(Option.DONT_CREATE_ID3_TAGS)) ? "" : null);
                else if (param == Id3.v2)
                    oPane.setValue(param, (!options.containsKey(Option.NO_ID3V2) && !options
                            .containsKey(Option.DONT_CREATE_ID3_TAGS)) ? "" : null);

                // check whether oPane differs from default settings
                if (!isDefault) continue;
                if (param == Id3.v1)
                    for (Option o : new Option[]{Option.DONT_CREATE_ID3_TAGS, Option.NO_ID3V2, Option.ADD_ID3V1})
                        if (!isDefault) break;
                        else isDefault = defaultOptions.containsKey(o) == options.containsKey(o);
                else if (param == OverwriteMode.always)
                    param = Option.OVERWRITE_COMPLETE_TRACKS;
                if (param instanceof Option)
                    isDefault = (defaultOptions.containsKey(param) == options.containsKey(param))
                            && (defaultOptions.get(param) == null ? options.get(param) == null
                                    : defaultOptions.get(param).equals(options.get(param)));
            }

            //switch expanded state only if options differ from default settings
            oPane.setExpanded(defaultOptions!=null && !isDefault);
        }
    }

    public Map<Option, String> getStreamRipperOptions()
    {
        Map<Option, String> options= new TreeMap<Option, String>();
        if (defaultOptions != null) options.putAll(defaultOptions);
        options.put(Option.STREAM_URL, null);
        options.put(Option.DESTINATION_DIRECTORY, null);
        options.put(Option.CREATE_RELAY_STREAM, null);

        for(ExpandableOptionPane oPane : optionPanes)
            if (oPane.isExpanded())
            {
                for(Enum param : oPane.getParams())
                    if (!oPane.isSet(param)) continue;
                    else if (param instanceof Option)
                    {
                        String value = oPane.getValue(param);

                        // switch values for Option.TRACK_PADDING
                        if (param == Option.TRACK_PADDING && value!=null && value.contains(":"))
                        {
                            String[] splitted = value.split(":");
                            value = splitted[1]+":"+splitted[0];
                        }

                        options.put((Option) param, value);
                    }
                    else if (param instanceof OverwriteMode)
                        options.put(Option.OVERWRITE_COMPLETE_TRACKS, oPane.getValue(param));
            }

        // id3-tag param to option mapping
        boolean id3v1 = id3tags.isSet(Id3.v1), 
                id3v2 = id3tags.isSet(Id3.v2);
        if (!id3tags.isExpanded());
        else if (!id3v2 && !id3v1) options.put(Option.DONT_CREATE_ID3_TAGS, null);
        else
        {
            if (id3v1) options.put(Option.ADD_ID3V1, null);
            if (!id3v2) options.put(Option.NO_ID3V2, null);
        }

        return options;
    }

    public void writeToBaseModel()
    {
        model.setValue(Config.Key.STREAMRIPPER_OPTS, StreamripperCommand
                .compileOptionString(getStreamRipperOptions()));
        model.setValue(Config.Key.STREAMRIPPER_RELAY_PORT_RANGE, getRelayPortRange());
    }

    public void readFromBaseModel()
    {
        try
        {
            setStreamRipperOptions(StreamripperCommand.decompileOptionString(model
                    .getValue(Key.STREAMRIPPER_OPTS)));
            setRelayPortRange(model.getValue(Key.STREAMRIPPER_RELAY_PORT_RANGE));
        }
        catch (RuntimeException e)
        {
            // TODO handle case that optionstring is not valid
            // show dialog which is saying that we start over from default settings
            e.printStackTrace();
            setStreamRipperOptions(defaultOptions);
        }
    }

    @Override
    public void applyChanges()
    {
        writeToBaseModel();
        for (Key param : configParams) model.apply(param);
        readFromBaseModel();
    }

    @Override
    public void loadDefaults()
    {
        for (Key param : configParams) model.restore(param);
        readFromBaseModel();
    }

    @Override
    public void revertChanges()
    {
        for (Key param : configParams) model.revert(param);
        readFromBaseModel();
    }
}