/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.gui.views;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.CancellationException;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.SwingWorker;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.TableColumnModel;

import org.kuehntopf.streamtastic.core.DirectoryFeed;
import org.kuehntopf.streamtastic.core.Station;
import org.kuehntopf.streamtastic.core.config.Config;
import org.kuehntopf.streamtastic.gui.StreamtasticFrame;
import org.kuehntopf.streamtastic.gui.components.ColumnSelectionPopupMenu;
import org.kuehntopf.streamtastic.gui.components.DirectoryPanel;
import org.kuehntopf.streamtastic.gui.components.StationColumn;
import org.kuehntopf.streamtastic.gui.components.StationPopupMenu;
import org.kuehntopf.streamtastic.gui.models.BitrateListModel;
import org.kuehntopf.streamtastic.gui.models.GenreListModel;
import org.kuehntopf.streamtastic.gui.models.StationTableModel;
import org.kuehntopf.streamtastic.utils.Utilities;

/**
 * Shoutcast View
 * @author  Andreas Kuehntopf
 */
public class DirectoryView extends ToolbarView
{
    private final static int SOURCE_GENRE=0;
    private final static int SOURCE_SEARCH=1;
        
    private DirectoryFeed feed;
    private StreamtasticFrame sframe;
    private boolean busy = false;
    private DirectoryPanel directoryPanel;
    private GenreListModel genreModel;
    private List<SwingWorker<?, ?>> runningWorkers = new Vector<SwingWorker<?,?>>(5);

    public DirectoryView(StreamtasticFrame parent, DirectoryFeed feed)
    {
        super(feed.getName(), feed.getImageIcons()[DirectoryFeed.ICON_LARGE],
                feed.getImageIcons()[DirectoryFeed.ICON_SMALL]);
        this.feed = feed;
        this.sframe = parent;

        // abort running workers on shutdown
        sframe.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e)
            {
                for (SwingWorker<?, ?> worker : runningWorkers
                        .toArray(new SwingWorker<?, ?>[runningWorkers.size()]))
                    worker.cancel(true);
            }
        });

        init();
        updateButtons();
    }

    @Override
    protected JComponent initContentPane()
    {
        directoryPanel = new DirectoryPanel();

        // Setup Station table and Header Popup 
        TableColumnModel columnModel = new DefaultTableColumnModel();
        for (Station.Property type :feed.getSupportedStationProperties()) 
            columnModel.addColumn(new StationColumn(type));
        ColumnSelectionPopupMenu headermenu = new ColumnSelectionPopupMenu(columnModel);
        for (Station.Property property : new Station.Property[] { Station.Property.GENRE, Station.Property.URL })
            headermenu.setSelected(property,false);

        final StationTableModel stationTableModel = new StationTableModel();
        final JTable stationTable = directoryPanel.initTable(stationTableModel, columnModel,
                StationColumn.getRowsorter(stationTableModel));
        stationTable.getTableHeader().setComponentPopupMenu(headermenu);
        stationTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent evt)
            {
                updateButtons();
            }
        });

        // configure double click action
        stationTable.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e)
            {
                if (e.getClickCount() != 2)
                    return;

                Action action = sframe.getSharedAction(StreamtasticFrame.StationAction
                        .valueOf(Config.getInstance().getProperty(
                                Config.Key.DIR_STATION_DOUBLE_CLICK_ACTION)));

                if (action != null && action.isEnabled())
                    action.actionPerformed(new ActionEvent(e.getSource(),
                            ActionEvent.ACTION_PERFORMED, StreamtasticFrame.StationAction.RECORD
                                    .toString(), e.getWhen(), e.getModifiers()));
            }
        });

        // add popup menu
        JPopupMenu tblmenu = new StationPopupMenu();
        tblmenu.add(sframe.getSharedAction(StreamtasticFrame.StationAction.PLAY));
        tblmenu.add(sframe.getSharedAction(StreamtasticFrame.StationAction.RECORD));
        tblmenu.add(sframe.getSharedAction(StreamtasticFrame.StationAction.ABORT));
        tblmenu.add(sframe.getSharedAction(StreamtasticFrame.StationAction.ADD_TO_FAVS));
        tblmenu.addSeparator();
        tblmenu.add(sframe.getSharedAction(StreamtasticFrame.StationAction.COPY_URL_TO_CLIPBOARD));
        stationTable.setComponentPopupMenu(tblmenu);

        // GenreList
        genreModel = new GenreListModel();
        ListSelectionListener listener = new ListSelectionListener() 
        {
            private String previousGenre = null;

            public void valueChanged(ListSelectionEvent evt)
            {
                if (evt.getValueIsAdjusting()) return;

                JList listGenres = (JList) evt.getSource();

                // do nothing if current genre is null or equal to previous genre
                // may happen, when current selected genre was moved during a genre list refresh
                // or the genre list selection was cleared due to a search request
                String currentGenre = (String) listGenres.getSelectedValue();
                if (currentGenre == null || currentGenre.equals(previousGenre)) 
                {
                    previousGenre = currentGenre;
                    return;
                }

                //clear the selection, if selected genre changed
                stationTable.clearSelection();
                
                refreshStationTable();                
                updateButtons();
                previousGenre = currentGenre;
            }
        };
        directoryPanel.initGenreList(genreModel, listener);

        directoryPanel.initBitrates(new BitrateListModel(stationTableModel), new ActionListener() {
            public void actionPerformed(ActionEvent e)
            {
                refreshStationTable();
            }
        });

        Action searchAction = new AbstractAction() {
            public void actionPerformed(ActionEvent e)
            {
                JTextField tf = (JTextField) e.getSource();
                
                directoryPanel.clearGenreSelection();
                refreshStationTable(tf.getText(), SOURCE_SEARCH);
            }
        };
        directoryPanel.setSearchAction(searchAction);
        
        return directoryPanel;
    }

    @Override
    protected void initToolbar(JToolBar toolbar)
    {
        toolbar.add(sframe.getSharedAction(StreamtasticFrame.StationAction.PLAY));
        toolbar.add(sframe.getSharedAction(StreamtasticFrame.StationAction.RECORD));
        toolbar.add(sframe.getSharedAction(StreamtasticFrame.StationAction.ABORT));
        toolbar.add(sframe.getSharedAction(StreamtasticFrame.StationAction.ADD_TO_FAVS));

        JButton btnRefresh = new JButton();
        btnRefresh.setIcon(new javax.swing.ImageIcon(getClass().getResource("/org/kuehntopf/streamtastic/resources/icons/small/refresh.png"))); // NOI18N
        btnRefresh.setToolTipText(Utilities.getString("REFRESH")); // NOI18N
        btnRefresh.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                feed.flush();
                refreshGenreList(genreModel.getSize() > 0);
                refreshStationTable();
            }
        });
        toolbar.add(btnRefresh);
    }

    private void setBusy(boolean busy)
    {
        this.busy = busy;
        sframe.setBusyPaneVisibleForView(this, busy);
    }

    private boolean isBusy()
    {
        return busy;
    }

    private void updateButtons()
    {
        sframe.updateStationActions(directoryPanel.getSelectedStation());
    }

    /**
     * Gets called when the View is activated
     */
    @Override
    protected void activate()
    {
        if (isBusy()) sframe.setBusyPaneVisibleForView(this, true);

        if (!isBusy() && !isTaskWaitingForExecution() && genreModel.getSize() == 0)
            refreshGenreList(false);

        updateButtons();
    }

    private void refreshGenreList(final boolean inBackground)
    {
        SwingWorker<List<String>, Void> worker = new SwingWorker<List<String>, Void>() {
            @Override
            protected List<String> doInBackground() throws Exception
            {
                return feed.getGenres();
            }

            @Override
            protected void done()
            {
                runningWorkers.remove(this);

                if (!inBackground)
                    setBusy(false);
                try
                {
                    directoryPanel.setGenres(get());
                } catch (CancellationException ce)
                {
                    // should only happen during shutdown
                    System.err.println(feed.getName() + " genre retrieval canceled!");
                }
                catch (Exception e)
                {
                    if (!inBackground)
                        executeWhenActive(new Runnable() { public void run() {
                            if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(sframe,
                                        Utilities.getString("DIR_GENRELIST_ERROR_MESSAGE"),
                                        Utilities.getString("DIR_GENRELIST_ERROR_TITLE"),
                                        JOptionPane.YES_NO_OPTION, JOptionPane.ERROR_MESSAGE))
                                    activate();
                        }});
                }
            }
        };

        if (!inBackground)
            setBusy(true);
        runningWorkers.add(worker);
        worker.execute();    
    }
    
    private void refreshStationTable()
    {
        if (directoryPanel.getSelectedGenre() != null)
        {
            refreshStationTable(directoryPanel.getSelectedGenre(), SOURCE_GENRE);
            directoryPanel.clearSearchTerm();
        } else if (directoryPanel.getSearchTerm().length() > 2)
            refreshStationTable(directoryPanel.getSearchTerm(), SOURCE_SEARCH);
    }
    
    private void refreshStationTable(final String query, final int source)
    {              
        if (source==SOURCE_SEARCH && query.length()<3)
        {
            JOptionPane.showMessageDialog(sframe, 
                    Utilities.getString("DIR_SEARCH_ERROR_MESSAGE"),
                    Utilities.getString("DIR_STATIONLIST_ERROR_TITLE"), 
                    JOptionPane.ERROR_MESSAGE);
            directoryPanel.clearStations();
            return;
        }
                
        SwingWorker<List<Station>, Void> worker = new SwingWorker<List<Station>, Void>() 
        {
            @Override
            protected List<Station> doInBackground() throws Exception 
            {
                Integer bitrate = directoryPanel.getSelectedBitrate();
                
                switch (source)
                {
                    case SOURCE_GENRE: 
                        // retrieve new stations
                        return feed.getStations(query,bitrate);
                    case SOURCE_SEARCH:
                        return feed.getSearchResult(query,bitrate);
                    default:
                      return null;
                }
            }

            @Override
            protected void done() 
            {
                runningWorkers.remove(this);

                setBusy(false);
                try
                {
                    directoryPanel.setStations(get());
                } catch (CancellationException ce)
                {
                    // should only happen during shutdown
                    System.err.println(feed.getName() + " station retrieval canceled!");
                }
                catch (Exception e)
                {
                    executeWhenActive(new Runnable() { public void run() {
                        if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(sframe, Utilities
                                .getString("DIR_STATIONLIST_ERROR_MESSAGE"), Utilities
                                .getString("DIR_STATIONLIST_ERROR_TITLE"), JOptionPane.YES_NO_OPTION,
                                JOptionPane.ERROR_MESSAGE))
                            refreshStationTable(query, source);

                        //clear station table
                        directoryPanel.clearStations();
                    }});
                }
            }
        };
        setBusy(true);
        runningWorkers.add(worker);
        worker.execute(); 
    }
}
