/*
 * Streamtastic Streamripper Frontend
 * Copyright 2008 Andreas Kuehntopf <andreas@kuehntopf.org>
 * Copyright 2008-2010 Michael Lippold <thabird@free.fr>
 * Copyright 2008-2011 Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 * 
 * This file is part of Streamtastic.
 *
 * Streamtastic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Streamtastic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Streamtastic.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.kuehntopf.streamtastic.utils;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeMap;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author Joachim Lippold <mailToJoeAtIcq@icqmail.com>
 *
 */
public class Command {

    private static final Pattern parameterPattern = Pattern.compile(".*\\$\\{([\\w-_=]+)\\}.*");
    private static Command COMMAND;

    private ProcessBuilder processBuilder;
    private List<String> command;
    private Map<String, ParameterData> parameters = new TreeMap<String, ParameterData>();

    public Command(String executable, String options) throws FileNotFoundException
    {
        this(executable, options, true);
    }

    protected Command(String executable, String options, boolean checkCanExecute)
            throws FileNotFoundException
    {
        if (checkCanExecute)
        {
            File execFile = new File(executable);
            if (!execFile.canExecute())
                throw new FileNotFoundException("Cannot execute " + execFile.getPath());
        }
        if (executable != null)
            processBuilder = new ProcessBuilder(executable);

        command = new Vector<String>();
        command.add(executable);
        if (options != null)
            command.addAll(getOptionsList(options));
    }

    public void setOptionString(String options)
    {
        String exec = command.get(0);
        command.clear();
        command.add(exec);
        parameters.clear();
        if (options!=null) command.addAll(getOptionsList(options));
    }

    public void set(String parameter, String value) throws IllegalArgumentException
    {
        if (!parameters.containsKey(parameter))
            throw new IllegalArgumentException("Parameter \"" + parameter + "\" not defined!");

        parameters.get(parameter).value = value;
    }

    public Process execute() throws IOException, IllegalStateException
    {
        return processBuilder.command(replaceParameters()).start();
    }

    public String[] getParameters()
    {
        return parameters.keySet().toArray(new String[0]);
    }

    protected Process execute(String... params) throws IOException, IllegalStateException
    {
        List<String> command = new Vector<String>();
        command.add(this.command.get(0));
        command.addAll(Arrays.asList(params));
        return processBuilder.command(command).start();
    }

    static protected String replaceParameter(String optionString, String parameterName, String value)
    {
        if (COMMAND==null) try {
            COMMAND= new Command(null, null, false);
        } catch (FileNotFoundException e) { return ""; /*never happens*/ };

        COMMAND.setOptionString(optionString);
        COMMAND.set(parameterName, value);

        return COMMAND.toString();
    }

    static protected String[] decomposeOptionString(String optionString)
    {
        if (COMMAND==null) try {
            COMMAND= new Command(null, null, false);
        } catch (FileNotFoundException e) { return new String[0]; /*never happens*/ };

        COMMAND.setOptionString(optionString);
        return COMMAND.command.toArray(new String[COMMAND.command.size()]);
    }

    @Override
    public String toString()
    {
        StringBuffer cmd = new StringBuffer();
        for (String s: replaceParameters())
        {
            if (cmd.length()!=0) cmd.append(" ");
            cmd.append(s.contains(" ")?("'"+s+"'"):s);
        }
        return cmd.toString();
    }

    private String[] replaceParameters()
    {
        Vector<String> cmd = new Vector<String>(command.size());
        cmd.addAll(command);
        for (String parameter : parameters.keySet())
        {
            ParameterData p = parameters.get(parameter);
            if (p.value == null)
                throw new IllegalStateException("There is no value for \"" + parameter + "\"!");
            else
            {
                if (p.value.length() != 0)
                {
                    for (int i = 0; i < p.prepend.length; i++)
                        cmd.set(p.idx - p.prepend.length + i, p.prepend[i]);
                    cmd.set(p.idx, cmd.get(p.idx).replace("${" + parameter + "}", p.value));
                }
                else cmd.set(p.idx, null);
            }
        }
        cmd.removeAll(Arrays.asList(new String[] { null, "" }));
        // for (String opt: cmd) System.out.print(opt+" ");
        // System.out.println();
        return cmd.toArray(new String[cmd.size()]);
    }

    private List<String> getOptionsList(String options)
    {
        // command line args with spaces must be escaped with single quotes
        // single quotes on unix work similar to double quotes on win32
        // double quotes are not accepted because escaping is handled
        // differently on unix and win32
        if (options != null && options.contains("\""))
            throw new IllegalArgumentException(
                    "[\"] not allowed in options string! Use single quotes instead!");

        List<String> optionList = new Vector<String>();
        Scanner scanner = new Scanner(options);
        scanner.useDelimiter("\\s+");
        String cmdOption;
        Matcher parameterMatcher;
        while (scanner.hasNext())
        {
            if (scanner.hasNext("[^']*['].*"))
            // a quote is starting in next token
            {
                cmdOption = scanner.findWithinHorizon("([^'\\s]*['][^']*['])+[^'\\s]*", 0);
                if (cmdOption == null)
                    throw new IllegalArgumentException("Unclosed quote encountered in \"" + options
                            + "\"!");
                cmdOption = cmdOption.replaceAll("'", "");
            }
            else
                cmdOption = scanner.next();

            if ((parameterMatcher = parameterPattern.matcher(cmdOption)).matches())
            {
                ParameterData p = new ParameterData(parameterMatcher.group(1),
                        optionList.size() + 1);
                parameters.put(p.param, p);
                cmdOption = cmdOption.replace(parameterMatcher.group(1), p.param);
                for (int i = 0; i < p.prepend.length; i++)
                    optionList.add(null);
            }
            optionList.add(cmdOption);
        }

        return optionList;
    }

}

class ParameterData {
    protected int idx;
    protected String value;
    protected String param;
    protected String[] prepend;

    protected ParameterData(String parameter, int idx)
    {
        String[] options = parameter.split("_");
        param = options[options.length - 1];
        prepend = new String[options.length - 1];
        this.idx = idx + prepend.length;
        for (int i = 0; i < prepend.length; i++)
            prepend[i] = options[i];
    }
}
